"""
Tests for the pager module.
"""

import unittest

from flask import current_app, Flask

from kerko import blueprint as kerko_blueprint
from kerko import pager


class SectionsTestCase(unittest.TestCase):
    """Test pager sections when there is a single page."""

    def setUp(self):
        self.app = Flask(__name__)
        self.app.register_blueprint(kerko_blueprint, url_prefix='/bibliography')
        ctx = self.app.app_context()
        ctx.push()

    def test_default_config(self):
        # This setting impacts the results of pager functions.
        self.assertEqual(current_app.config['KERKO_PAGER_LINKS'], 4)

    def test_on_page_1_of_1(self):
        sections = pager.get_sections(page_num=1, page_count=1)
        self.assertIsNone(sections)
        pages = pager.get_page_numbers(sections)
        self.assertFalse(pages)

    def test_on_page_1_of_2(self):
        sections = pager.get_sections(page_num=1, page_count=2)
        self.assertNotIn('previous', sections)
        self.assertNotIn('first', sections)
        self.assertFalse(sections['before'])
        self.assertEqual(sections['current'], [1])
        self.assertEqual(sections['after'], [2])
        self.assertEqual(sections['last'], [2])
        self.assertEqual(sections['next'], [2])
        pages = pager.get_page_numbers(sections)
        self.assertEqual(pages, [1, 2])

    def test_on_page_2_of_2(self):
        sections = pager.get_sections(page_num=2, page_count=2)
        self.assertEqual(sections['previous'], [1])
        self.assertEqual(sections['first'], [1])
        self.assertEqual(sections['before'], [1])
        self.assertEqual(sections['current'], [2])
        self.assertFalse(sections['after'])
        self.assertNotIn('last', sections)
        self.assertNotIn('next', sections)
        pages = pager.get_page_numbers(sections)
        self.assertEqual(pages, [1, 2])

    def test_on_page_3_of_5(self):
        sections = pager.get_sections(page_num=3, page_count=5)
        self.assertEqual(sections['previous'], [2])
        self.assertEqual(sections['first'], [1])
        self.assertEqual(sections['before'], [1, 2])
        self.assertEqual(sections['current'], [3])
        self.assertEqual(sections['after'], [4, 5])
        self.assertEqual(sections['last'], [5])
        self.assertEqual(sections['next'], [4])
        pages = pager.get_page_numbers(sections)
        self.assertEqual(pages, [1, 2, 3, 4, 5])

    def test_on_page_5_of_9(self):
        sections = pager.get_sections(page_num=5, page_count=9)
        self.assertEqual(sections['previous'], [4])
        self.assertEqual(sections['first'], [1])
        self.assertEqual(sections['before'], [3, 4])
        self.assertEqual(sections['current'], [5])
        self.assertEqual(sections['after'], [6, 7])
        self.assertEqual(sections['last'], [9])
        self.assertEqual(sections['next'], [6])
        pages = pager.get_page_numbers(sections)
        self.assertEqual(pages, [1, 3, 4, 5, 6, 7, 9])

    def test_on_page_999_of_1000(self):
        sections = pager.get_sections(page_num=999, page_count=1000)
        self.assertEqual(sections['previous'], [998])
        self.assertEqual(sections['first'], [1])
        self.assertEqual(sections['before'], [996, 997, 998])
        self.assertEqual(sections['current'], [999])
        self.assertEqual(sections['after'], [1000])
        self.assertEqual(sections['last'], [1000])
        self.assertEqual(sections['next'], [1000])
        pages = pager.get_page_numbers(sections)
        self.assertEqual(pages, [1, 996, 997, 998, 999, 1000])

    def test_min_page_num(self):
        sections = pager.get_sections(page_num=-1, page_count=2)
        self.assertNotIn('previous', sections)
        self.assertNotIn('first', sections)
        self.assertFalse(sections['before'])
        self.assertEqual(sections['current'], [1])
        self.assertEqual(sections['after'], [2])
        self.assertEqual(sections['last'], [2])
        self.assertEqual(sections['next'], [2])
        pages = pager.get_page_numbers(sections)
        self.assertEqual(pages, [1, 2])

    def test_max_page_num(self):
        sections = pager.get_sections(page_num=3, page_count=2)
        self.assertEqual(sections['previous'], [1])
        self.assertEqual(sections['first'], [1])
        self.assertEqual(sections['before'], [1])
        self.assertEqual(sections['current'], [2])
        self.assertFalse(sections['after'])
        self.assertNotIn('last', sections)
        self.assertNotIn('next', sections)
        pages = pager.get_page_numbers(sections)
        self.assertEqual(pages, [1, 2])

    def test_invalid_page_count(self):
        sections = pager.get_sections(page_num=2, page_count=-1)
        self.assertIsNone(sections)


if __name__ == '__main__':
    unittest.main()



"""
Custom filters for Jinja2 templates.
"""

from .datetime import format_datetime, reformat_date


def register_filters(blueprint):
    """Add custom template filters."""
    blueprint.add_app_template_filter(format_datetime, name='kerko_format_datetime')
    blueprint.add_app_template_filter(reformat_date, name='kerko_reformat_date')



import copy
import sys
import time
from datetime import datetime

from babel.dates import format_datetime
from babel.numbers import format_number
from flask import (abort, current_app, flash, make_response, redirect,
                   render_template, request, send_from_directory, url_for)
from flask_babel import get_locale, get_timezone, gettext, ngettext

from . import babel_domain, blueprint
from .attachments import get_attachments_dir
from .breadbox import build_breadbox
from .criteria import Criteria
from .forms import SearchForm
from .pager import build_pager, get_page_numbers
from .pager import get_sections as get_pager_sections
from .query import (build_creators_display, build_item_facet_results,
                    build_relations, get_search_return_fields, run_query,
                    run_query_unique_with_fallback)
from .sorter import build_sorter

if sys.version_info < (3, 7):
    # Workaround for 'TypeError: cannot deepcopy this pattern object' when
    # encountering a compiled regular expression.
    import re
    copy._deepcopy_dispatch[type(re.compile(''))] = lambda r, _: r  # pylint: disable=protected-access


@blueprint.route('/', methods=['GET', 'POST'])
def search():
    """View the results of a search."""
    start_time = time.process_time()

    if current_app.config['KERKO_USE_TRANSLATIONS']:
        babel_domain.as_default()

    criteria = Criteria(request)
    form = SearchForm(csrf_enabled=False)
    if form.validate_on_submit():
        url = criteria.build_add_keywords_url(
            scope=form.scope.data,
            value=form.keywords.data)
        return redirect(url, 302)

    search_results, facet_results, total_count, page_count, last_sync = run_query(
        criteria, get_search_return_fields(criteria.page_len)
    )

    if criteria.page_len == 1 and criteria.id and (
            total_count == 0 or criteria.id != search_results[0]['id']
    ):
        # The search result page no longer points to the desired item.
        return redirect(url_for('.item_view', item_id=criteria.id, _external=True), 301)

    criteria.fit_pager(page_count)
    breadbox = build_breadbox(criteria, facet_results)
    pager_sections = get_pager_sections(criteria.page_num, page_count)
    context = {
        'facet_results': facet_results,
        'breadbox': breadbox,
        'active_facets': breadbox['filters'].keys() if 'filters' in breadbox else [],
        'sorter': build_sorter(criteria),
        'total_count': total_count,
        'total_count_formatted': format_number(total_count, locale=get_locale()),
        'page_count': page_count,
        'page_count_formatted': format_number(page_count, locale=get_locale()),
        'page_len': criteria.page_len,
        'show_abstracts': criteria.show_abstracts,
        'abstracts_toggler_url': criteria.build_url(
            show_abstracts=not criteria.show_abstracts, page_num=criteria.page_num
        ),
        'is_searching': criteria.has_keyword_search() or criteria.has_filter_search(),
        'locale': get_locale(),
        'last_sync': datetime.fromtimestamp(
            last_sync, tz=datetime.now().astimezone().tzinfo
        ) if last_sync else None,
    }

    if criteria.page_len == 1 and total_count != 0:
        # Retrieve item ids corresponding to individual result page numbers.
        page_kwargs = {}
        page_criteria = copy.deepcopy(criteria)
        for page_num in get_page_numbers(pager_sections):
            if page_num == criteria.page_num:
                # We already know the current page's item id. No further query necessary.
                page_kwargs[page_num] = {'id_': search_results[0]['id']}
            else:
                # Run a search query to get the item id corresponding to the page number.
                page_criteria.page_num = page_num
                page_search_results, _, _, _, _ = run_query(
                    page_criteria, return_fields=['id'], query_facets=False
                )
                if page_search_results:
                    page_kwargs[page_num] = {'id_': page_search_results[0]['id']}
        context['pager'] = build_pager(pager_sections, criteria, page_kwargs)

        list_page_num = int((criteria.page_num - 1) / current_app.config['KERKO_PAGE_LEN'] + 1)
        build_creators_display(search_results[0])
        build_item_facet_results(search_results[0])
        build_relations(
            search_results[0],
            get_search_return_fields(page_len=None, exclude=['coins']),
            sort=current_app.config['KERKO_RELATIONS_SORT']
        )
        if context['is_searching']:
            context['search_title'] = gettext('Your search')
        else:
            context['search_title'] = gettext('Full bibliography')
        return render_template(
            current_app.config['KERKO_TEMPLATE_SEARCH_ITEM'],
            title=search_results[0].get('data', {}).get('title', ''),
            item=search_results[0],
            item_url=url_for(
                '.item_view', item_id=search_results[0]['id'], _external=True
            ) if search_results[0] else '',
            back_url=criteria.build_url(page_num=list_page_num),
            time=time.process_time() - start_time,
            **context
        )

    if total_count > 0:
        context['pager'] = build_pager(pager_sections, criteria)
        search_results_urls = [
            criteria.build_url(
                page_num=(criteria.page_num - 1) * (criteria.page_len or 0) + i + 1,
                page_len=1,
                id_=result['id'],
            ) for i, result in enumerate(search_results)
        ]
        search_results = zip(search_results, search_results_urls)
        if context['is_searching']:
            context['title'] = ngettext('Result', 'Results', total_count)
        else:
            context['title'] = gettext('Full bibliography')
    else:
        context['title'] = gettext('Your search did not match any resources')
    return render_template(
        current_app.config['KERKO_TEMPLATE_SEARCH'],
        form=form,
        search_results=search_results,
        print_url=criteria.build_url(page_len='all', print_preview=True),
        print_preview=criteria.print_preview,
        download_urls={
            key: criteria.build_download_url(key)
            for key in current_app.config['KERKO_COMPOSER'].citation_formats.keys()
        },
        time=time.process_time() - start_time,
        **context
    )


@blueprint.route('/<path:item_id>')
def item_view(item_id):
    """View a full bibliographic record."""
    start_time = time.process_time()

    if current_app.config['KERKO_USE_TRANSLATIONS']:
        babel_domain.as_default()

    item, fellback = run_query_unique_with_fallback(['id', 'alternateId'], item_id)
    if not item:
        return abort(404)
    item_url = url_for('.item_view', item_id=item['id'], _external=True)
    if fellback:
        return redirect(item_url, 301)

    build_creators_display(item)
    build_item_facet_results(item)
    build_relations(
        item,
        get_search_return_fields(page_len=None, exclude=['coins']),
        sort=current_app.config['KERKO_RELATIONS_SORT']
    )
    return render_template(
        current_app.config['KERKO_TEMPLATE_ITEM'],
        item=item,
        title=item.get('data', {}).get('title', ''),
        item_url=item_url,
        time=time.process_time() - start_time,
        locale=get_locale(),
    )


@blueprint.route('/<path:item_id>/download/<string:attachment_id>/')
@blueprint.route('/<path:item_id>/download/<string:attachment_id>/<string:attachment_filename>')
def item_attachment_download(item_id, attachment_id, attachment_filename=None):
    """
    Download an item attachment.

    If the URL does not specify the attachment's filename or provides the wrong
    filename, a redirect is performed to a corrected URL so that the client gets
    a proper filename.
    """
    if current_app.config['KERKO_USE_TRANSLATIONS']:
        babel_domain.as_default()

    item, fellback = run_query_unique_with_fallback(['id', 'alternateId'], item_id)
    if not item:
        return abort(404)

    matching_attachments = list(
        filter(lambda a: a.get('id') == attachment_id, item.get('attachments', []))
    )
    if not matching_attachments or len(matching_attachments) > 1:
        flash(
            gettext(
                "The document you have requested has been removed. "
                "Please check below for the latest documents available."
            ), 'warning'
        )
        return redirect(url_for('.item_view', item_id=item['id']), 301)
    attachment = matching_attachments[0]

    filepath = get_attachments_dir() / attachment_id
    if not filepath.exists():
        return abort(404)

    if fellback or attachment_filename != attachment['filename']:
        return redirect(url_for(
            '.item_attachment_download',
            item_id=item['id'],
            attachment_id=attachment_id,
            attachment_filename=attachment['filename'],
        ), 301)

    return send_from_directory(
        get_attachments_dir(),
        attachment_id,
        mimetype=attachment['mimetype'],
    )


@blueprint.route('/<path:item_id>/export/<string:citation_format_key>')
def item_citation_download(item_id, citation_format_key):
    """Download a citation."""
    if current_app.config['KERKO_USE_TRANSLATIONS']:
        babel_domain.as_default()

    item, fellback = run_query_unique_with_fallback(['id', 'alternateId'], item_id)
    if not item:
        return abort(404)

    citation_format = current_app.config['KERKO_COMPOSER'].citation_formats.get(citation_format_key)
    if not citation_format:
        return abort(404)

    content = item.get(citation_format.field.key)
    if not content:
        return abort(404)

    if fellback:
        return redirect(
            url_for(
                '.item_citation_download',
                item_id=item['id'],
                citation_format_key=citation_format_key
            ), 301
        )

    response = make_response(content)
    response.headers['Content-Disposition'] = \
        f"attachment; filename={item['id']}.{citation_format.extension}"
    response.headers['Content-Type'] = \
        f'{citation_format.mime_type}; charset=utf-8'
    return response


@blueprint.route('/export/<string:citation_format_key>/')
def search_citation_download(citation_format_key):
    """Download all citations resulting from a search."""
    citation_format = current_app.config['KERKO_COMPOSER'].citation_formats.get(citation_format_key)
    if not citation_format:
        return abort(404)

    criteria = Criteria(request)
    criteria.page_len = None

    search_results, _, total_count, _, _ = run_query(  # TODO: Avoid building facet results.
        criteria, return_fields=[citation_format.field.key]
    )

    if total_count == 0:
        return abort(404)

    citations = [result.get(citation_format.field.key, '') for result in search_results]
    response = make_response(
        citation_format.group_format.format(
            citation_format.group_item_delimiter.join(citations)
        )
    )
    response.headers['Content-Disposition'] = \
        f'attachment; filename=bibliography.{citation_format.extension}'  # TODO: Make filename configurable.
    response.headers['Content-Type'] = \
        f'{citation_format.mime_type}; charset=utf-8'
    return response



import pprint
from datetime import datetime

import click
import wrapt
from flask import current_app
from flask.cli import with_appcontext

from . import zotero
from .attachments import delete_attachments, sync_attachments
from .index import delete_index, sync_index


@wrapt.decorator
def execution_time_logger(wrapped, _instance, args, kwargs):
    start_time = datetime.now()
    return_value = wrapped(*args, **kwargs)
    current_app.logger.info(_format_elapsed_time(start_time))
    return return_value


@click.group()
def cli():
    """Run a Kerko subcommand."""


@cli.command(deprecated=True)
@with_appcontext
def index():  # Deprecated after version 0.4.
    """
    Synchronize the search index and the attachments from the Zotero library.

    This command is deprecated. Use the 'sync' command.
    """
    sync_index()
    sync_attachments()


@cli.command()
@click.argument(
    'target',
    default='everything',
    type=click.Choice(['index', 'attachments', 'everything'], case_sensitive=False),
)
@with_appcontext
@execution_time_logger
def sync(target):
    """
    Synchronize the search index and/or the attachments from the Zotero library.

    By default, everything is synchronized.
    """
    if target in ['everything', 'index']:
        sync_index()
    if target in ['everything', 'attachments']:
        sync_attachments()


@cli.command()
@click.argument(
    'target',
    default='everything',
    type=click.Choice(['index', 'attachments', 'everything'], case_sensitive=False),
)
@with_appcontext
def clean(target):
    """
    Delete the search index and/or the attachments immediately.

    By default, both are cleaned.
    """
    if target in ['everything', 'index']:
        delete_index()
    if target in ['everything', 'attachments']:
        delete_attachments()


@cli.command()
@click.argument('item_key')
@with_appcontext
def zotero_item(item_key):
    """
    Retrieve an item from the Zotero library.

    WARNING: This command is provided for debugging purposes only and may be
    modified or removed from the module at any time.
    """
    credentials = zotero.init_zotero()
    pprint.pprint(zotero.load_item(credentials, item_key))


@cli.command()
@with_appcontext
def zotero_item_types():
    """
    List all item types supported by Zotero.

    WARNING: This command is provided for debugging purposes only and may be
    modified or removed from the module at any time.
    """
    credentials = zotero.init_zotero()
    pprint.pprint(zotero.load_item_types(credentials))


@cli.command()
@with_appcontext
def zotero_item_fields():
    """
    List all fields supported by Zotero.

    WARNING: This command is provided for debugging purposes only and may be
    modified or removed from the module at any time.
    """
    credentials = zotero.init_zotero()
    pprint.pprint(zotero.load_item_fields(credentials))


@cli.command()
@click.argument('item_type')
@with_appcontext
def zotero_item_type_fields(item_type):
    """
    List all fields supported by Zotero for a given item type.

    WARNING: This command is provided for debugging purposes only and may be
    modified or removed from the module at any time.
    """
    credentials = zotero.init_zotero()
    pprint.pprint(zotero.load_item_type_fields(credentials, item_type))


@cli.command()
@click.argument('item_type')
@with_appcontext
def zotero_item_type_creator_types(item_type):
    """
    List all creator types supported by Zotero for a given item type.

    WARNING: This command is provided for debugging purposes only and may be
    modified or removed from the module at any time.
    """
    credentials = zotero.init_zotero()
    pprint.pprint(zotero.load_item_type_creator_types(credentials, item_type))


@cli.command()
@with_appcontext
def zotero_top_level_collections():
    """
    List keys and names of top-level collections from the Zotero library.

    WARNING: This command is provided for debugging purposes only and may be
    modified or removed from the module at any time.
    """
    credentials = zotero.init_zotero()
    collections = zotero.Collections(credentials, top_level=True)
    for c in collections:
        print(f"{c.get('key')} {c.get('data', {}).get('name', '')}")


def _format_elapsed_time(start_time):
    elapsed_time = int(round((datetime.now() - start_time).total_seconds()))
    elapsed_min, elapsed_sec = elapsed_time // 60, elapsed_time % 60
    s = 'Execution time:'
    if elapsed_min > 0:
        s += (' {n} minutes' if elapsed_min > 1 else ' {n} minute').format(n=elapsed_min)
        s += (' {n:02} seconds' if elapsed_sec > 1 else ' {n:02d} second').format(n=elapsed_sec)
    else:
        s += (' {n} seconds' if elapsed_sec > 1 else ' {n} second').format(n=elapsed_sec)
    return s



import hashlib
import pathlib

from flask import current_app

from . import zotero
from .query import check_fields, run_query_all


def get_attachments_dir():
    return pathlib.Path(current_app.config['KERKO_DATA_DIR']) / 'attachments'


def md5_checksum(path):
    # Hash file in 8k blocks to avoid reading it all in memory.
    with path.open('rb') as f:
        md5_hash = hashlib.md5()
        while True:
            data = f.read(8192)
            if not data:
                break
            md5_hash.update(data)
        return md5_hash.hexdigest()


def sync_attachments():
    """
    Synchronize attachments from Zotero into the attachments directory.

    Files are requested based on item data available in the search index. Thus,
    it always makes sense to synchronize the search index beforehand.
    """
    current_app.logger.info("Starting attachments sync.")
    attachments_dir = get_attachments_dir()
    attachments_dir.mkdir(parents=True, exist_ok=True)
    local_files = {p.name for p in attachments_dir.iterdir()}

    missing_fields = check_fields(['id', 'attachments'])
    if missing_fields:
        current_app.logger.error(
            "The following fields are missing from the search index: {}."
            " You might need to check your Kerko settings and/or"
            " clean and sync your search index.".format(
                ', '.join(missing_fields)
            )
        )
        return

    # List all items from the search index and request their attachments, if
    # any, from Zotero
    zotero_credentials = zotero.init_zotero()
    count = 0
    parents = run_query_all(['id', 'attachments'])
    for parent in parents:
        for attachment in parent.get('attachments', []):
            if not attachment['id']:
                current_app.logger.warning(
                    f"A child attachment of {parent['id']} lacks an id. Skipped."
                )
                break
            if not attachment['md5']:
                current_app.logger.warning(
                    f"A child attachment of {parent['id']} lacks a checksum."
                )
            if attachment['id'] in local_files:
                local_files.remove(attachment['id'])
            filepath = attachments_dir / attachment['id']
            if not filepath.exists() or md5_checksum(filepath) != attachment['md5']:
                current_app.logger.debug(
                    f"Requesting attachment {attachment['id']} (parent: {parent['id']})..."
                )
                try:
                    # Download attachment.
                    with filepath.open('wb') as f:
                        f.write(zotero.retrieve_file(zotero_credentials, attachment['id']))
                except zotero.zotero_errors.PyZoteroError as e:
                    current_app.logger.exception(
                        f"Unexpected error while attempting to download attachment "
                        f"{attachment['id']} (parent: {parent['id']}): {e}"
                    )
            else:
                current_app.logger.debug(
                    f"Keeping attachment {attachment['id']} (parent: {parent['id']})."
                )
            count += 1

    # Delete remaining local files that were not referenced by any item.
    for name in local_files:
        current_app.logger.debug(f"Deleting attachment {name}, unused.")
        (attachments_dir / name).unlink()

    current_app.logger.info(f"Attachments sync successful ({count} attachment(s) processed).")
    return count


def delete_attachments():
    attachments_dir = get_attachments_dir()
    if attachments_dir.is_dir():
        for filepath in attachments_dir.iterdir():
            filepath.unlink()
        attachments_dir.rmdir()



from itertools import chain

from babel.numbers import format_number
from flask import current_app
from flask_babel import get_locale


def get_sections(page_num, page_count):
    """Return dict of pager sections, where each section is a list of page numbers."""
    if page_count <= 1:
        return None

    sections = {}
    page_num = min(max(page_num, 1), page_count)
    first = max(1, int(page_num - current_app.config['KERKO_PAGER_LINKS'] / 2))
    last = min(page_count + 1, first + current_app.config['KERKO_PAGER_LINKS'] + 1)
    first = max(1, last - current_app.config['KERKO_PAGER_LINKS'] - 1)

    if page_num > 1:
        sections['first'] = [1]
        sections['previous'] = [max(1, page_num - 1)]

    sections['before'] = list(range(first, page_num))
    sections['current'] = [page_num]
    sections['after'] = list(range(page_num + 1, last))

    if page_num < page_count:
        sections['next'] = [min(page_count, page_num + 1)]
        sections['last'] = [page_count]

    return sections


def get_page_numbers(sections):
    """Return distinct pages numbers used across pager sections."""
    if sections:
        return sorted(set(chain(*sections.values())))
    return []


def build_pager(sections, criteria, page_kwargs=None):
    """
    Build pager links for use in a search view.

    :param dict page_kwargs: A dict keyed by page number, where each value is a
        dict of extra params to pass along when building that page's URL.
    """
    pager = {}
    if sections:
        for section, pages in sections.items():
            pager[section] = [
                {
                    'page_num': p,
                    'page_num_formatted': format_number(p, locale=get_locale()),
                    'url': criteria.build_url(
                        page_num=p,
                        page_len=criteria.page_len,
                        **page_kwargs.get(p) if page_kwargs else {}
                    )
                } for p in pages
            ]
    return pager



"""
Tests for the date and time utilities.
"""

import unittest

from flask import Flask, current_app
from flask_babel import Babel, Domain
from kerko import blueprint as kerko_blueprint
from kerko.datetime import reformat_date


class DateStringReformattingTestCase(unittest.TestCase):
    """Test the date string reformatting function."""

    def setUp(self):
        self.app = Flask(__name__)
        self.app.register_blueprint(kerko_blueprint, url_prefix='/bibliography')
        babel_domain = Domain()
        babel = Babel(default_domain=babel_domain)
        babel.init_app(self.app)
        ctx = self.app.app_context()
        ctx.push()

    def test_non_iso8601_utc_en(self):
        current_app.config['BABEL_DEFAULT_LOCALE'] = 'en_US'
        current_app.config['BABEL_DEFAULT_TIMEZONE'] = 'UTC'
        self.assertEqual(reformat_date('2020'), '2020')
        self.assertEqual(reformat_date('2020-10'), '2020-10')
        self.assertEqual(reformat_date('2020-05-05'), '2020-05-05')
        self.assertEqual(reformat_date('01/02/2021'), '01/02/2021')
        self.assertEqual(reformat_date('September 2020'), 'September 2020')
        self.assertEqual(reformat_date('September 20, 2020'), 'September 20, 2020')
        self.assertEqual(
            reformat_date('September 20, 2020', convert_tz=True),
            'September 20, 2020')
        self.assertEqual(
            reformat_date('September 20, 2020', convert_tz=True, show_tz=True),
            'September 20, 2020'
        )

    def test_iso8601_utc_en(self):
        current_app.config['BABEL_DEFAULT_LOCALE'] = 'en_US'
        current_app.config['BABEL_DEFAULT_TIMEZONE'] = 'UTC'
        self.assertEqual(
            reformat_date('2020-09-17T07:11:41+00:00', convert_tz=True),
            '9/17/20, 7:11 AM'
        )
        self.assertEqual(
            reformat_date('2020-09-17T07:11:41+00:00', convert_tz=False),
            '9/17/20, 7:11 AM'
        )
        self.assertEqual(
            reformat_date('2020-09-17T07:11:41+00:00', convert_tz=True, show_tz=True),
            '9/17/20, 7:11 AM (UTC)'
        )
        self.assertEqual(
            reformat_date('2020-11-07T12:30:19Z', convert_tz=True),
            '11/7/20, 12:30 PM'
        )
        self.assertEqual(
            reformat_date('2020-11-07T12:30:19Z', convert_tz=False),
            '11/7/20, 12:30 PM'
        )
        self.assertEqual(
            reformat_date('2020-11-07T12:30:19Z', convert_tz=True, show_tz=True),
            '11/7/20, 12:30 PM (UTC)'
        )
        self.assertEqual(
            reformat_date('2019-07-22T20:42+00:00', convert_tz=True),
            '7/22/19, 8:42 PM'
        )
        self.assertEqual(
            reformat_date('2019-07-22T20:42+00:00', convert_tz=False),
            '7/22/19, 8:42 PM'
        )
        self.assertEqual(
            reformat_date('2019-07-22T20:42+00:00', convert_tz=True, show_tz=True),
            '7/22/19, 8:42 PM (UTC)'
        )
        self.assertEqual(
            reformat_date('2020-11-07T12:30:19-0500', convert_tz=True),
            '11/7/20, 5:30 PM'
        )
        self.assertEqual(
            reformat_date('2020-11-07T12:30:19-0500', convert_tz=False),
            '11/7/20, 12:30 PM'
        )
        self.assertEqual(
            reformat_date('2020-11-07T12:30:19-0500', convert_tz=True, show_tz=True),
            '11/7/20, 5:30 PM (UTC)'
        )

    def test_iso8601_est_en(self):
        current_app.config['BABEL_DEFAULT_LOCALE'] = 'en_US'
        current_app.config['BABEL_DEFAULT_TIMEZONE'] = 'EST'
        self.assertEqual(
            reformat_date('2020-09-17T07:11:41+00:00', convert_tz=True),
            '9/17/20, 2:11 AM'
        )
        self.assertEqual(
            reformat_date('2020-09-17T07:11:41+00:00', convert_tz=False),
            '9/17/20, 7:11 AM'
        )
        self.assertEqual(
            reformat_date('2020-09-17T07:11:41+00:00', convert_tz=True, show_tz=True),
            '9/17/20, 2:11 AM (EST)'
        )
        self.assertEqual(
            reformat_date('2020-11-07T12:30:19Z', convert_tz=True),
            '11/7/20, 7:30 AM'
        )
        self.assertEqual(
            reformat_date('2020-11-07T12:30:19Z', convert_tz=False),
            '11/7/20, 12:30 PM'
        )
        self.assertEqual(
            reformat_date('2020-11-07T12:30:19Z', convert_tz=True, show_tz=True),
            '11/7/20, 7:30 AM (EST)'
        )
        self.assertEqual(
            reformat_date('2019-07-22T20:42+00:00', convert_tz=True),
            '7/22/19, 3:42 PM'
        )
        self.assertEqual(
            reformat_date('2019-07-22T20:42+00:00', convert_tz=False),
            '7/22/19, 8:42 PM'
        )
        self.assertEqual(
            reformat_date('2019-07-22T20:42+00:00', convert_tz=True, show_tz=True),
            '7/22/19, 3:42 PM (EST)'
        )
        self.assertEqual(
            reformat_date('2020-11-07T12:30:19-0500', convert_tz=True),
            '11/7/20, 12:30 PM'
        )
        self.assertEqual(
            reformat_date('2020-11-07T12:30:19-0500', convert_tz=False),
            '11/7/20, 12:30 PM'
        )
        self.assertEqual(
            reformat_date('2020-11-07T12:30:19-0500', convert_tz=True, show_tz=True),
            '11/7/20, 12:30 PM (EST)'
        )

    def test_non_iso8601_utc_fr(self):
        current_app.config['BABEL_DEFAULT_LOCALE'] = 'fr_FR'
        current_app.config['BABEL_DEFAULT_TIMEZONE'] = 'UTC'
        self.assertEqual(reformat_date('September 2020'), 'September 2020')
        self.assertEqual(reformat_date('September 20, 2020'), 'September 20, 2020')

    def test_iso8601_utc_fr(self):
        current_app.config['BABEL_DEFAULT_LOCALE'] = 'fr_FR'
        current_app.config['BABEL_DEFAULT_TIMEZONE'] = 'UTC'
        self.assertEqual(
            reformat_date('2020-09-17T07:11:41+00:00', convert_tz=True),
            '17/09/2020 07:11'
        )
        self.assertEqual(
            reformat_date('2020-09-17T07:11:41+00:00', convert_tz=False),
            '17/09/2020 07:11'
        )
        self.assertEqual(
            reformat_date('2020-09-17T07:11:41+00:00', convert_tz=True, show_tz=True),
            '17/09/2020 07:11 (UTC)'
        )
        self.assertEqual(
            reformat_date('2020-11-07T12:30:19Z', convert_tz=True),
            '07/11/2020 12:30'
        )
        self.assertEqual(
            reformat_date('2020-11-07T12:30:19Z', convert_tz=False),
            '07/11/2020 12:30'
        )
        self.assertEqual(
            reformat_date('2020-11-07T12:30:19Z', convert_tz=True, show_tz=True),
            '07/11/2020 12:30 (UTC)'
        )
        self.assertEqual(
            reformat_date('2019-07-22T20:42+00:00', convert_tz=True),
            '22/07/2019 20:42'
        )
        self.assertEqual(
            reformat_date('2019-07-22T20:42+00:00', convert_tz=False),
            '22/07/2019 20:42'
        )
        self.assertEqual(
            reformat_date('2019-07-22T20:42+00:00', convert_tz=True, show_tz=True),
            '22/07/2019 20:42 (UTC)'
        )
        self.assertEqual(
            reformat_date('2020-11-07T12:30:19-0500', convert_tz=True),
            '07/11/2020 17:30'
        )
        self.assertEqual(
            reformat_date('2020-11-07T12:30:19-0500', convert_tz=False),
            '07/11/2020 12:30'
        )
        self.assertEqual(
            reformat_date('2020-11-07T12:30:19-0500', convert_tz=True, show_tz=True),
            '07/11/2020 17:30 (UTC)'
        )



import pathlib
import shutil

from flask import current_app
import whoosh

from . import zotero
from .extractors import ItemContext, LibraryContext
from .tags import TagGate


def get_index_dir():
    return pathlib.Path(current_app.config['KERKO_DATA_DIR']) / 'index'


def open_index(auto_create=False, write=False):
    index_dir = get_index_dir()
    try:
        if not index_dir.exists() and auto_create and write:
            index_dir.mkdir(parents=True, exist_ok=True)
            return whoosh.index.create_in(
                str(index_dir), current_app.config['KERKO_COMPOSER'].schema
            )
        if index_dir.exists():
            return whoosh.index.open_dir(str(index_dir), readonly=not write)
        current_app.logger.error(f"Could not open index in directory '{index_dir}'.")
    except whoosh.index.IndexError as e:
        current_app.logger.error(f"Could not open index: '{e}'.")


def delete_index():
    index_dir = get_index_dir()
    if index_dir.is_dir():
        shutil.rmtree(index_dir)


def request_library_context(zotero_credentials):
    item_types = {
        t['itemType']: t['localized']
        for t in zotero.load_item_types(zotero_credentials)
    }
    return LibraryContext(
        zotero_credentials.library_id,
        zotero_credentials.library_type.rstrip('s'),  # Remove 's' appended by pyzotero.
        collections=zotero.Collections(zotero_credentials),
        item_types=item_types,
        item_fields={
            t: zotero.load_item_type_fields(zotero_credentials, t)
            for t in item_types.keys()
        },
        creator_types={
            t: zotero.load_item_type_creator_types(zotero_credentials, t)
            for t in item_types.keys()
        }
    )


def sync_index():
    """Build the search index from items retrieved from Zotero."""
    current_app.logger.info("Starting index sync.")
    composer = current_app.config['KERKO_COMPOSER']
    zotero_credentials = zotero.init_zotero()
    library_context = request_library_context(zotero_credentials)
    index = open_index(auto_create=True, write=True)
    count = 0

    def get_children(item):
        children = []
        if item.get('meta', {}).get('numChildren', 0):
            # TODO: Only extract item types that are required, if any, by the Composer instance's fields.
            children = list(
                zotero.ChildItems(
                    zotero_credentials,
                    item['key'],
                    item_types=['note', 'attachment'],
                )
            )
        return children

    writer = index.writer(limitmb=256)
    try:
        writer.mergetype = whoosh.writing.CLEAR
        allowed_item_types = [
            t for t in library_context.item_types.keys()
            if t not in ['note', 'attachment']
        ]
        formats = {
            spec.extractor.format
            for spec in list(composer.fields.values()) + list(composer.facets.values())
        }
        gate = TagGate(composer.default_item_include_re, composer.default_item_exclude_re)
        for item in zotero.Items(zotero_credentials, allowed_item_types, list(formats)):
            count += 1
            if gate.check(item.get('data', {})):
                item_context = ItemContext(item, get_children(item))
                document = {}
                for spec in list(composer.fields.values()) + list(composer.facets.values()):
                    spec.extract_to_document(document, item_context, library_context)
                update_document_with_writer(writer, document, count=count)
            else:
                current_app.logger.debug(f"Document {count} excluded ({item['key']})")
    except Exception as e:  # pylint: disable=broad-except
        writer.cancel()
        current_app.logger.exception(e)
        current_app.logger.error('An exception occurred. Could not finish updating the index.')
    else:
        writer.commit()
        current_app.logger.info(f"Index sync successful ({count} item(s) processed).")
    return count


def update_document_with_writer(writer, document, count=None):
    """
    Update a document in the search index.

    :param writer: The index writer.

    :param document: A dict whose fields match the schema.

    :param count: An optional document count, for logging purposes.
    """
    writer.update_document(**document)
    current_app.logger.debug(
        'Document {count}updated ({id}): {title}'.format(
            id=document.get('id', ''),
            title=document.get('z_title'),
            count='' if count is None else '{} '.format(count)
        )
    )


def update_document(document):
    """
    Update a document in the search index.

    :param document: A dict whose fields match the schema.
    """
    index = open_index(write=True)
    with index.writer(limitmb=256) as writer:
        update_document_with_writer(writer, document)



import re  # pylint: disable=too-many-lines

import whoosh
from flask_babel import lazy_gettext as _
from whoosh.analysis import CharsetFilter, LowercaseFilter, StemFilter
from whoosh.analysis.tokenizers import RegexTokenizer
from whoosh.fields import BOOLEAN, COLUMN, ID, NUMERIC, STORED, TEXT, Schema, columns
from whoosh.query import Prefix, Term
from whoosh.support.charset import accent_map
from whoosh.util.text import rcompile

from . import codecs, extractors, transformers
from .specs import (CitationFormatSpec, FieldSpec, FlatFacetSpec, RelationSpec, ScopeSpec, SortSpec,
                    TreeFacetSpec)


class Composer:
    """
    A factory for the setting up the search elements.

    This class acts as a registry for configuring the search elements such as
    fields and facets, from which both the schema and the search interface can
    be built.

    The schema is the representation of documents and their fields in the search
    index. It is meant to be fully constructed at app configuration time and not
    changed afterwards. If schema elements need to be modified or removed, the
    application should be stopped, and the search index cleaned and rebuilt.
    """

    def __init__(
            self,
            *,
            whoosh_language='en',
            exclude_default_scopes=None,
            exclude_default_fields=None,
            exclude_default_facets=None,
            exclude_default_sorts=None,
            exclude_default_citation_formats=None,
            exclude_default_relations=None,
            exclude_default_badges=None,
            default_item_include_re='',
            default_item_exclude_re='',
            default_tag_include_re='',
            default_tag_exclude_re=r'^_',
            default_child_include_re='',
            default_child_exclude_re=r'^_',
            mime_types=None,
    ):  # pylint: disable=too-many-arguments
        """
        Initialize the object.

        Based on the provided arguments, default fields and facets may be
        initialized. Those may be modified after initialization; it is safe to
        change attributes that impact the search interface and not the schema,
        but those changes that should impact the schema require the field to be
        removed first, then re-added with the changes.

        :param list exclude_default_scopes: List of scopes (identified by key)
            to exclude from those created by default. If that list contains the
            value '*', no scope will be added by default. Caution: most default
            fields are expecting one or more of those scopes to exist. You'll
            have to add them manually or alter the fields. Please refer to the
            implementation of ``init_default_scopes()`` for the list of default
            scopes.

        :param list exclude_default_fields: List of fields (identified by key)
            to exclude from those created by default. If that list contains the
            value '*', no field will be created by default. Caution: some
            default fields are required by Kerko. If those are excluded, you'll
            have to add them manually. Please refer to the implementation of
            ``init_default_fields()`` for the list of default fields.

        :param list exclude_default_facets: List of facets (identified by key)
            to exclude from those created by default. If that list contains the
            value '*', no facet will be created by default. Please refer to the
            implementation of ``init_default_facets()`` for the list of default
            facets.

        :param list exclude_default_sorts: List of sorts (identified by key) to
            exclude from those created by default. If that list contains the
            value '*', no sort will be created by default and you'll have to
            manually add at least one sort. Please refer to the implementation
            of ``init_default_sorts()`` for the list of default sorts.

        :param list exclude_default_citation_formats: List of citation download
            formats (identified by key) to exclude from those created by
            default. If that list contains the value '*', no citation format
            will be created by default. Please refer to the implementation of
            ``init_default_citation_formats()`` for the list of default citation
            formats.

        :param list exclude_default_relations: List of relation types
            (identified by key) to exclude from those created by default. If
            that list contains the value '*', no relation type will be created
            by default. Please refer to the implementation of
            ``init_default_relations()`` for the list of default relation types.

        :param list exclude_default_badges: List of badges (identified by key)
            to exclude from those created by default. If that list contains the
            value '*', no badge will be created by default. Please refer to the
            implementation of ``init_default_badges()`` for the list of default
            badges.

        :param str default_item_include_re: Regex to use to include items based
            on their tags. Any item which does not have a tag that matches this
            regular expression will be excluded. If empty (which is the default),
            all items will be included unless the `default_item_exclude_re`
            argument is set and causes some to be excluded.

        :param str default_item_exclude_re: Regex to use to exclude items based
            on their tags. Any item that have a tag that matches this regular
            expression will be excluded. If empty (which is the default), no
            items will be excluded unless the `default_item_include_re` argument
            is set, in which case items that don't have any tag that matches it
            will be excluded.

        :param str default_tag_include_re: Regex to use to include tags. See
            ``extractors.BaseTagsExtractor``. By default, all tags are accepted.
            Note that citation exports (downloads) always include all tags
            regardless of this parameter, which only applies to information
            displayed by Kerko (exports are generated by the Zotero API, not by
            Kerko).

        :param str default_tag_exclude_re: Regex to use to exclude tags. See
            ``extractors.BaseTagsExtractor``. The default value causes any tag
            that begins with an underscore ('_') to be ignored by Kerko. Note
            that citation exports (downloads) always include all tags regardless
            of this parameter, which only applies to information displayed by
            Kerko (exports are generated by the Zotero API, not by Kerko).

        :param str default_child_include_re: Regex to use to include
            children (e.g. notes, attachments) based on their tags. Any child
            which does not have a tag that matches this regular expression will
            be ignored by the extractor. If empty, all children will be accepted
            unless the `default_child_exclude_re` argument is set and causes
            some to be rejected.

        :param str default_child_exclude_re: Regex to use to exclude
            children (e.g. notes, attachments) based on their tags. Any child
            that have a tag that matches this regular expression will be ignored
            by the extractor. If empty, no children will be rejected unless the
            `default_child_include_re` argument is set and the tags of those
            children don't match it. By default, any child having at least one
            tag that begins with an underscore ('_') is rejected.

        :param list mime_types: List of allowed MIME types for attachments. If
            this parameter is None, the list will default to
            `["application/pdf"]`. Pass an empty list to allow any MIME type.
        """
        if not whoosh.lang.has_stemmer(whoosh_language):
            whoosh_language = 'en'

        # When combining multiple strings into a single text field, the last
        # token of each string becomes adjacent to the first token of the next
        # string. This may cause phrase searches to match those tokens as if
        # they were neighbors, even though they were not in the source data. To
        # prevent this, we join the strings with the record separator character
        # and treat that character as a token. This solution is imperfect,
        # however, as the issue may still arise when a slop factor is applied to
        # the phrase search.
        token_pattern = rcompile(r"\w+(\.?\w+)*|" + re.escape(extractors.RECORD_SEPARATOR))

        # Replace the standard analyzer with one that has no stop words (helping
        # people who do phrase searches without specifying actual phrase queries).
        self.text_chain = \
            RegexTokenizer(expression=token_pattern) | \
            StemFilter(lang=whoosh_language) | \
            CharsetFilter(accent_map) | \
            LowercaseFilter()

        # Same for names, but without stemming.
        self.name_chain = \
            RegexTokenizer(expression=token_pattern) | \
            CharsetFilter(accent_map) | \
            LowercaseFilter()

        # Common schema field types.
        self.id_field_type = ID(field_boost=50.0)
        self.secondary_id_field_type = ID(field_boost=10.0)
        self.title_field_type = TEXT(analyzer=self.text_chain, field_boost=5.0)
        self.secondary_title_field_type = TEXT(analyzer=self.text_chain, field_boost=2.0)
        self.creator_name_field_type = TEXT(analyzer=self.name_chain, field_boost=2.0)
        self.name_field_type = TEXT(analyzer=self.name_chain)
        self.text_field_type = TEXT(analyzer=self.text_chain)

        self.schema = Schema()
        self.scopes = {}
        self.fields = {}
        self.facets = {}
        self.sorts = {}
        self.citation_formats = {}
        self.relations = {}
        self.badges = {}
        self.default_item_include_re = default_item_include_re
        self.default_item_exclude_re = default_item_exclude_re
        self.default_tag_include_re = default_tag_include_re
        self.default_tag_exclude_re = default_tag_exclude_re
        self.default_child_include_re = default_child_include_re
        self.default_child_exclude_re = default_child_exclude_re
        if mime_types is None:
            self.mime_types = ['application/pdf']
        else:
            self.mime_types = mime_types
        self.init_default_scopes(exclude_default_scopes)
        self.init_default_fields(exclude_default_fields)
        self.init_default_facets(exclude_default_facets)
        self.init_default_sorts(exclude_default_sorts)
        self.init_default_citation_formats(exclude_default_citation_formats)
        self.init_default_relations(exclude_default_relations)
        self.init_default_badges(exclude_default_badges)

    def init_default_scopes(self, exclude=None):
        if exclude is None:
            exclude = []

        if '*' in exclude:
            return

        if 'all' not in exclude:
            self.add_scope(
                ScopeSpec(
                    key='all',
                    selector_label=_("In any field"),
                    breadbox_label=_("Any field"),
                    weight=0,
                    help_text=_(
                        "Finds entries where any of the available fields contains your keywords. "
                        "This is the default option."
                    ),
                )
            )
        if 'creator' not in exclude:
            self.add_scope(
                ScopeSpec(
                    key='creator',
                    selector_label=_("In authors/contributors"),
                    breadbox_label=_("Author or contributor"),
                    weight=100,
                    help_text=_(
                        "Finds entries where any of its authors' or contributors' names contains "
                        "your keywords."
                    ),
                )
            )
        if 'title' not in exclude:
            self.add_scope(
                ScopeSpec(
                    key='title',
                    selector_label=_("In titles"),
                    breadbox_label=_("Title"),
                    weight=200,
                    help_text=_("Finds entries whose title contains your keywords."),
                )
            )

    def init_default_fields(self, exclude=None):
        if exclude is None:
            exclude = []

        if '*' in exclude:
            return

        # Primary item ID for the search index, matching the Zotero item key.
        if 'id' not in exclude:
            self.add_field(
                FieldSpec(
                    key='id',
                    field_type=ID(unique=True, stored=True),
                    extractor=extractors.ItemExtractor(key='key')
                )
            )
        # Alternate IDs to search when the primary ID is not found.
        if 'alternateId' not in exclude:
            self.add_field(
                FieldSpec(
                    key='alternateId',
                    field_type=ID(stored=True),
                    extractor=extractors.MultiExtractor(
                        extractors=[
                            extractors.ItemDataExtractor(key='DOI'),
                            extractors.ItemDataExtractor(key='ISBN'),
                            extractors.ItemDataExtractor(key='ISSN'),
                            # Extract DOI, ISBN, ISSN from the Extra field.
                            extractors.TransformerExtractor(
                                extractor=extractors.ItemDataExtractor(key='extra'),
                                transformers=[
                                    transformers.find(
                                        regex=r'^\s*(DOI|ISBN|ISSN):\s*(\S+)\s*$',
                                        flags=re.IGNORECASE | re.MULTILINE,
                                        group=2,
                                        max_matches=0,
                                    ),
                                ]
                            ),
                            # Extract dc:replaces relations.
                            extractors.TransformerExtractor(
                                extractor=extractors.ItemRelationsExtractor(
                                    predicate='dc:replaces'
                                ),
                                transformers=[transformers.find_item_id_in_zotero_uris_list]
                            )
                        ]
                    )
                )
            )
        # Label of this item's type.
        if 'item_type' not in exclude:
            self.add_field(
                FieldSpec(
                    key='item_type',
                    field_type=TEXT(analyzer=self.text_chain, stored=True),
                    scopes=['all'],
                    extractor=extractors.ItemTypeLabelExtractor(),
                )
            )

        # All Zotero item fields that we want to make available to keyword
        # search are specified below, each with an appropriate analyzer chain.
        # The full list of item fields can be obtained through the Zotero API.
        # Keys derive from the corresponding field names in Zotero, but are
        # prefixed with 'z_' to prevent name clashes with Kerko's own fields.

        #
        # Identifier fields, for keyword search. Highly boosted for matches to
        # rank at the top of results.
        #

        if 'z_DOI' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_DOI',
                    field_type=self.id_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='DOI')
                )
            )
        if 'z_ISBN' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_ISBN',
                    field_type=self.id_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='ISBN')
                )
            )
        if 'z_ISSN' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_ISSN',
                    field_type=self.id_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='ISSN')
                )
            )

        #
        # Secondary identifiers, for keyword search. Moderately boosted.
        #

        if 'z_applicationNumber' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_applicationNumber',
                    field_type=self.secondary_id_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='applicationNumber')
                )
            )
        if 'z_billNumber' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_billNumber',
                    field_type=self.secondary_id_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='billNumber')
                )
            )
        if 'z_callNumber' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_callNumber',
                    field_type=self.secondary_id_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='callNumber')
                )
            )
        if 'z_codeNumber' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_codeNumber',
                    field_type=self.secondary_id_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='codeNumber')
                )
            )
        if 'z_docketNumber' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_docketNumber',
                    field_type=self.secondary_id_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='docketNumber')
                )
            )
        if 'z_documentNumber' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_documentNumber',
                    field_type=self.secondary_id_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='documentNumber')
                )
            )
        if 'z_patentNumber' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_patentNumber',
                    field_type=self.secondary_id_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='patentNumber')
                )
            )
        if 'z_priorityNumbers' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_priorityNumbers',
                    field_type=self.secondary_id_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='priorityNumbers')
                )
            )
        if 'z_publicLawNumber' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_publicLawNumber',
                    field_type=self.secondary_id_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='publicLawNumber')
                )
            )
        if 'z_reportNumber' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_reportNumber',
                    field_type=self.secondary_id_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='reportNumber')
                )
            )

        #
        # Title fields, for keyword search.
        #

        if 'z_nameOfAct' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_nameOfAct',
                    field_type=self.title_field_type,
                    scopes=['all', 'title'],
                    extractor=extractors.ItemDataExtractor(key='nameOfAct')
                )
            )
        if 'z_shortTitle' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_shortTitle',
                    field_type=self.title_field_type,
                    scopes=['all', 'title'],
                    extractor=extractors.ItemDataExtractor(key='shortTitle')
                )
            )
        if 'z_subject' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_subject',
                    field_type=self.title_field_type,
                    scopes=['all', 'title'],
                    extractor=extractors.ItemDataExtractor(key='subject')
                )
            )
        if 'z_title' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_title',
                    field_type=self.title_field_type,
                    scopes=['all', 'title'],
                    extractor=extractors.ItemDataExtractor(key='title')
                )
            )

        #
        # Container titles, for keyword search.
        #

        if 'z_blogTitle' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_blogTitle',
                    field_type=self.secondary_title_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='blogTitle')
                )
            )
        if 'z_bookTitle' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_bookTitle',
                    field_type=self.secondary_title_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='bookTitle')
                )
            )
        if 'z_code' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_code',
                    field_type=self.secondary_title_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='code')
                )
            )
        if 'z_conferenceName' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_conferenceName',
                    field_type=self.secondary_title_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='conferenceName')
                )
            )
        if 'z_dictionaryTitle' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_dictionaryTitle',
                    field_type=self.secondary_title_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='dictionaryTitle')
                )
            )
        if 'z_encyclopediaTitle' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_encyclopediaTitle',
                    field_type=self.secondary_title_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='encyclopediaTitle')
                )
            )
        if 'z_forumTitle' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_forumTitle',
                    field_type=self.secondary_title_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='forumTitle')
                )
            )
        if 'z_meetingName' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_meetingName',
                    field_type=self.secondary_title_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='meetingName')
                )
            )
        if 'z_proceedingsTitle' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_proceedingsTitle',
                    field_type=self.secondary_title_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='proceedingsTitle')
                )
            )
        if 'z_programTitle' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_programTitle',
                    field_type=self.secondary_title_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='programTitle')
                )
            )
        if 'z_publicationTitle' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_publicationTitle',
                    field_type=self.secondary_title_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='publicationTitle')
                )
            )
        if 'z_section' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_section',
                    field_type=self.secondary_title_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='section')
                )
            )
        if 'z_series' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_series',
                    field_type=self.secondary_title_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='series')
                )
            )
        if 'z_seriesTitle' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_seriesTitle',
                    field_type=self.secondary_title_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='seriesTitle')
                )
            )
        if 'z_session' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_session',
                    field_type=self.secondary_title_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='session')
                )
            )
        if 'z_websiteTitle' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_websiteTitle',
                    field_type=self.secondary_title_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='websiteTitle')
                )
            )

        #
        # Name fields, for keyword search. Exempt from stemming.
        #

        if 'z_archive' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_archive',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='archive')
                )
            )
        if 'z_archiveLocation' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_archiveLocation',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='archiveLocation')
                )
            )
        if 'z_assignee' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_assignee',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='assignee')
                )
            )
        if 'z_audioFileType' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_audioFileType',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='audioFileType')
                )
            )
        if 'z_audioRecordingFormat' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_audioRecordingFormat',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='audioRecordingFormat')
                )
            )
        if 'z_caseName' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_caseName',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='caseName')
                )
            )
        if 'z_committee' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_committee',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='committee')
                )
            )
        if 'z_company' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_company',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='company')
                )
            )
        if 'z_country' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_country',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='country')
                )
            )
        if 'z_court' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_court',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='court')
                )
            )
        if 'z_distributor' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_distributor',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='distributor')
                )
            )
        if 'z_institution' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_institution',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='institution')
                )
            )
        if 'z_issuingAuthority' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_issuingAuthority',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='issuingAuthority')
                )
            )
        if 'z_journalAbbreviation' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_journalAbbreviation',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='journalAbbreviation')
                )
            )
        if 'z_label' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_label',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='label')
                )
            )
        if 'z_legislativeBody' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_legislativeBody',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='legislativeBody')
                )
            )
        if 'z_libraryCatalog' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_libraryCatalog',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='libraryCatalog')
                )
            )
        if 'z_network' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_network',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='network')
                )
            )
        if 'z_place' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_place',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='place')
                )
            )
        if 'z_programmingLanguage' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_programmingLanguage',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='programmingLanguage')
                )
            )
        if 'z_publisher' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_publisher',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='publisher')
                )
            )
        if 'z_reporter' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_reporter',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='reporter')
                )
            )
        if 'z_studio' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_studio',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='studio')
                )
            )
        if 'z_system' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_system',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='system')
                )
            )
        if 'z_university' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_university',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='university')
                )
            )
        if 'z_videoRecordingFormat' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_videoRecordingFormat',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='videoRecordingFormat')
                )
            )

        #
        # Date fields, for keyword search.
        #

        if 'z_dateDecided' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_dateDecided',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='dateDecided')
                )
            )
        if 'z_dateEnacted' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_dateEnacted',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='dateEnacted')
                )
            )
        if 'z_date' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_date',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='date')
                )
            )
        if 'z_filingDate' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_filingDate',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='filingDate')
                )
            )
        if 'z_issueDate' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_issueDate',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='issueDate')
                )
            )

        #
        # Text fields, for keyword search.
        #

        if 'z_abstractNote' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_abstractNote',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='abstractNote')
                )
            )
        if 'z_artworkMedium' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_artworkMedium',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='artworkMedium')
                )
            )
        if 'z_artworkSize' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_artworkSize',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='artworkSize')
                )
            )
        if 'z_codeVolume' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_codeVolume',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='codeVolume')
                )
            )
        if 'z_edition' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_edition',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='edition')
                )
            )
        if 'z_episodeNumber' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_episodeNumber',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='episodeNumber')
                )
            )
        if 'z_extra' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_extra',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='extra')
                )
            )
        if 'z_genre' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_genre',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='genre')
                )
            )
        if 'z_history' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_history',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='history')
                )
            )
        if 'z_interviewMedium' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_interviewMedium',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='interviewMedium')
                )
            )
        if 'z_issue' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_issue',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='issue')
                )
            )
        if 'z_language' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_language',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='language')
                )
            )
        if 'z_legalStatus' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_legalStatus',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='legalStatus')
                )
            )
        if 'z_letterType' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_letterType',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='letterType')
                )
            )
        if 'z_manuscriptType' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_manuscriptType',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='manuscriptType')
                )
            )
        if 'z_mapType' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_mapType',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='mapType')
                )
            )
        if 'z_postType' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_postType',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='postType')
                )
            )
        if 'z_presentationType' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_presentationType',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='presentationType')
                )
            )
        if 'z_references' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_references',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='references')
                )
            )
        if 'z_reporterVolume' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_reporterVolume',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='reporterVolume')
                )
            )
        if 'z_reportType' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_reportType',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='reportType')
                )
            )
        if 'z_rights' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_rights',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='rights')
                )
            )
        if 'z_seriesNumber' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_seriesNumber',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='seriesNumber')
                )
            )
        if 'z_seriesText' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_seriesText',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='seriesText')
                )
            )
        if 'z_thesisType' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_thesisType',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='thesisType')
                )
            )
        if 'z_versionNumber' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_versionNumber',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='versionNumber')
                )
            )
        if 'z_volume' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_volume',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='volume')
                )
            )
        if 'z_websiteType' not in exclude:
            self.add_field(
                FieldSpec(
                    key='z_websiteType',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor(key='websiteType')
                )
            )

        # Known Zotero item fields that we are deliberately ignoring for search:
        # - accessDate
        # - codePages
        # - extra
        # - firstPage
        # - numberOfVolumes
        # - numPages
        # - pages
        # - runningTime
        # - scale
        # - url

        #
        # Other metadata associated to Zotero items, for keyword search.
        #

        # Creators, exempt from stemming.
        if 'text_creator' not in exclude:
            self.add_field(
                FieldSpec(
                    key='text_creator',
                    field_type=self.creator_name_field_type,
                    scopes=['all', 'creator'],
                    extractor=extractors.CreatorsExtractor()
                )
            )
        if 'text_collections' not in exclude:
            self.add_field(
                FieldSpec(
                    key='text_collections',
                    field_type=self.secondary_title_field_type,
                    scopes=['all'],
                    extractor=extractors.CollectionNamesExtractor()
                )
            )
        if 'text_tags' not in exclude:
            self.add_field(
                FieldSpec(
                    key='text_tags',
                    field_type=self.secondary_title_field_type,
                    scopes=['all'],
                    extractor=extractors.TagsTextExtractor(
                        include_re=self.default_tag_include_re,
                        exclude_re=self.default_tag_exclude_re
                    )
                )
            )
        if 'text_notes' not in exclude:
            self.add_field(
                FieldSpec(
                    key='text_notes',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.NotesTextExtractor(
                        include_re=self.default_child_include_re,
                        exclude_re=self.default_child_exclude_re
                    )
                )
            )

        #
        # Relation fields, searchable for internal purposes only (hence the lack
        # of a 'scopes' parameter).
        #

        # References to items that are cited by the item.
        if 'rel_cites' not in exclude:
            self.add_field(
                FieldSpec(
                    key='rel_cites',
                    field_type=ID(stored=True),
                    extractor=extractors.RelationsInNotesExtractor(
                        include_re=r'_cites',
                        exclude_re=''
                    )
                )
            )
        # Items related through Zotero's relation field.
        if 'rel_related' not in exclude:
            self.add_field(
                FieldSpec(
                    key='rel_related',
                    field_type=ID(stored=True),
                    extractor=extractors.TransformerExtractor(
                        extractor=extractors.ItemRelationsExtractor(
                            predicate='dc:relation',
                        ),
                        transformers=[transformers.find_item_id_in_zotero_uris_list]
                    )
                )
            )

        #
        # Stored fields, not available for keyword search (hence the lack of a
        # 'scopes' parameter).
        #

        # Formatted citation.
        if 'bib' not in exclude:
            self.add_field(
                FieldSpec(
                    key='bib',
                    field_type=STORED,
                    extractor=extractors.ItemExtractor(key='bib', format_='bib')
                )
            )
        # OpenURL Coins.
        if 'coins' not in exclude:
            self.add_field(
                FieldSpec(
                    key='coins',
                    field_type=STORED,
                    extractor=extractors.ItemExtractor(key='coins', format_='coins')
                )
            )
        # RIS.
        if 'ris' not in exclude:
            self.add_field(
                FieldSpec(
                    key='ris',
                    field_type=STORED,
                    extractor=extractors.ItemExtractor(key='ris', format_='ris')
                )
            )
        # BibTeX.
        if 'bibtex' not in exclude:
            self.add_field(
                FieldSpec(
                    key='bibtex',
                    field_type=STORED,
                    extractor=extractors.ItemExtractor(key='bibtex', format_='bibtex')
                )
            )
        # Raw item data.
        if 'data' not in exclude:
            self.add_field(
                FieldSpec(
                    key='data',
                    field_type=STORED,
                    extractor=extractors.RawDataExtractor(),
                    codec=codecs.JSONFieldCodec()
                )
            )
        # Child notes of the item.
        if 'notes' not in exclude:
            self.add_field(
                FieldSpec(
                    key='notes',
                    field_type=STORED,
                    extractor=extractors.RawNotesExtractor(
                        include_re=self.default_child_include_re,
                        exclude_re=self.default_child_exclude_re
                    )
                )
            )
        # URL attachments of the item.
        if 'links' not in exclude:
            self.add_field(
                FieldSpec(
                    key='links',
                    field_type=STORED,
                    extractor=extractors.LinkedURIAttachmentsExtractor(
                        include_re=self.default_child_include_re,
                        exclude_re=self.default_child_exclude_re
                    )
                )
            )
        # File attachments of the item.
        if 'attachments' not in exclude:
            self.add_field(
                FieldSpec(
                    key='attachments',
                    field_type=STORED,
                    extractor=extractors.StoredFileAttachmentsExtractor(
                        mime_types=self.mime_types,
                        include_re=self.default_child_include_re,
                        exclude_re=self.default_child_exclude_re
                    )
                )
            )
        # Fields and labels for this item type, for convenient access.
        if 'item_fields' not in exclude:
            self.add_field(
                FieldSpec(
                    key='item_fields',
                    field_type=STORED,
                    extractor=extractors.ItemFieldsExtractor(),
                    codec=codecs.JSONFieldCodec()
                )
            )
        # Creator types for this item type, for convenient access.
        if 'creator_types' not in exclude:
            self.add_field(
                FieldSpec(
                    key='creator_types',
                    field_type=STORED,
                    extractor=extractors.CreatorTypesExtractor(),
                    codec=codecs.JSONFieldCodec()
                )
            )

        #
        # Fields for sorting.
        #

        if 'sort_title' not in exclude:
            self.add_field(
                FieldSpec(
                    key='sort_title',
                    field_type=COLUMN(columns.VarBytesColumn()),
                    extractor=extractors.SortItemDataExtractor(key='title'),
                )
            )
        if 'sort_creator' not in exclude:
            self.add_field(
                FieldSpec(
                    key='sort_creator',
                    field_type=COLUMN(columns.VarBytesColumn()),
                    extractor=extractors.SortCreatorExtractor(),
                )
            )
        if 'sort_date' not in exclude:
            self.add_field(
                FieldSpec(
                    key='sort_date',
                    field_type=NUMERIC,
                    extractor=extractors.SortDateExtractor(),
                )
            )

    def init_default_facets(self, exclude=None):
        if exclude is None:
            exclude = []

        if '*' in exclude:
            return

        if 'facet_tag' not in exclude:
            self.add_facet(
                FlatFacetSpec(
                    key='facet_tag',
                    title=_('Topic'),
                    filter_key='topic',
                    weight=100,
                    field_type=ID(stored=True),
                    extractor=extractors.TagsFacetExtractor(
                        include_re=self.default_tag_include_re,
                        exclude_re=self.default_tag_exclude_re
                    ),
                    codec=codecs.BaseFacetCodec(),
                    missing_label=None,
                    sort_key=['label'],
                    sort_reverse=False,
                    item_view=True,
                    allow_overlap=True,
                    query_class=Term
                )
            )
        if 'facet_item_type' not in exclude:
            self.add_facet(
                FlatFacetSpec(
                    key='facet_item_type',
                    title=_('Resource type'),
                    filter_key='type',
                    weight=200,
                    field_type=ID(stored=True),
                    extractor=extractors.ItemTypeFacetExtractor(),
                    codec=codecs.ItemTypeFacetCodec(),
                    missing_label=None,
                    sort_key=['label'],
                    sort_reverse=False,
                    item_view=False,
                    allow_overlap=False,
                    query_class=Prefix
                )
            )
        if 'facet_year' not in exclude:
            self.add_facet(
                TreeFacetSpec(
                    key='facet_year',
                    title=_('Publication year'),
                    filter_key='year',
                    weight=300,
                    field_type=ID(stored=True),
                    extractor=extractors.YearFacetExtractor(),
                    codec=codecs.YearTreeFacetCodec(),
                    missing_label=_('Unknown'),
                    sort_key=['label'],
                    sort_reverse=False,
                    item_view=False,
                    allow_overlap=True,
                    query_class=Prefix
                )
            )
        if 'facet_link' not in exclude:
            self.add_facet(
                FlatFacetSpec(
                    key='facet_link',
                    title=_('Online resource'),
                    filter_key='link',
                    weight=400,
                    field_type=BOOLEAN(stored=True),
                    extractor=extractors.ItemDataLinkFacetExtractor(key='url'),
                    codec=codecs.BooleanFacetCodec(),
                    missing_label=None,
                    sort_key=['label'],
                    sort_reverse=False,
                    item_view=False,
                    allow_overlap=False,
                    query_class=Term
                )
            )

    def init_default_sorts(self, exclude=None):
        """
        Initialize a set of default `SortSpec` instances.

        These rely on `FieldSpec` instances, which must have been added beforehand.
        """
        if exclude is None:
            exclude = []

        if '*' in exclude:
            return

        if 'score' not in exclude:
            self.add_sort(
                SortSpec(
                    key='score',
                    label=_('Relevance'),
                    weight=0,
                    fields=None,
                    # Sort by score is only possible on keyword search.
                    is_allowed=lambda criteria: criteria.has_keyword_search()
                )
            )
        if 'date_desc' not in exclude:
            self.add_sort(
                SortSpec(
                    key='date_desc',
                    label=_('Newest first'),
                    weight=10,
                    fields=[
                        self.fields['sort_date'],
                        self.fields['sort_creator'],
                        self.fields['sort_title']
                    ],
                    reverse=[
                        True,
                        False,
                        False,
                    ]
                )
            )
        if 'date_asc' not in exclude:
            self.add_sort(
                SortSpec(
                    key='date_asc',
                    label=_('Oldest first'),
                    weight=11,
                    fields=[
                        self.fields['sort_date'],
                        self.fields['sort_creator'],
                        self.fields['sort_title']
                    ]
                )
            )
        if 'author_asc' not in exclude:
            self.add_sort(
                SortSpec(
                    key='author_asc',
                    label=_('Author A-Z'),
                    weight=20,
                    fields=[
                        self.fields['sort_creator'],
                        self.fields['sort_title'],
                        self.fields['sort_date'],
                    ]
                )
            )
        if 'author_desc' not in exclude:
            self.add_sort(
                SortSpec(
                    key='author_desc',
                    label=_('Author Z-A'),
                    weight=21,
                    fields=[
                        self.fields['sort_creator'],
                        self.fields['sort_title'],
                        self.fields['sort_date'],
                    ],
                    reverse=[
                        True,
                        False,
                        False,
                    ],
                )
            )
        if 'title_asc' not in exclude:
            self.add_sort(
                SortSpec(
                    key='title_asc',
                    label=_('Title A-Z'),
                    weight=30,
                    fields=[
                        self.fields['sort_title'],
                        self.fields['sort_creator'],
                        self.fields['sort_date']
                    ]
                )
            )
        if 'title_desc' not in exclude:
            self.add_sort(
                SortSpec(
                    key='title_desc',
                    label=_('Title Z-A'),
                    weight=31,
                    fields=[
                        self.fields['sort_title'],
                        self.fields['sort_creator'],
                        self.fields['sort_date']
                    ],
                    reverse=[
                        True,
                        False,
                        False,
                    ],
                )
            )

    def init_default_citation_formats(self, exclude=None):
        """
        Initialize a set of default `CitationFormatSpec` instances.

        These rely on `FieldSpec` instances, which must have been added beforehand.
        """
        if exclude is None:
            exclude = []

        if '*' in exclude:
            return

        if 'ris' not in exclude:
            self.add_citation_format(
                CitationFormatSpec(
                    key='ris',
                    field=self.fields['ris'],
                    label=_("RIS"),
                    help_text=_("Recommended format for most reference management software"),
                    weight=10,
                    extension='ris',
                    mime_type='application/x-research-info-systems'
                )
            )
        if 'bibtex' not in exclude:
            self.add_citation_format(
                CitationFormatSpec(
                    key='bibtex',
                    field=self.fields['bibtex'],
                    label=_("BibTeX"),
                    help_text=_("Recommended format for BibTeX-specific software"),
                    weight=20,
                    extension='bib',
                    mime_type='application/x-bibtex'
                )
            )

    def init_default_relations(self, exclude=None):
        """
        Initialize a set of default `RelationSpec` instances.

        These rely on `FieldSpec` instances, which must have been added beforehand.
        """
        if exclude is None:
            exclude = []

        if '*' in exclude:
            return

        if 'cites' not in exclude:
            self.add_relation(
                RelationSpec(
                    key='cites',
                    field=self.fields['rel_cites'],
                    label=_("Cites"),
                    weight=10,
                    id_fields=[self.fields['id'], self.fields['alternateId']],
                    reverse=True,
                    reverse_key='isCitedBy',
                    reverse_field_key='rev_cites',
                    reverse_label=_("Cited by"),
                )
            )
        if 'related' not in exclude:
            self.add_relation(
                RelationSpec(
                    key='related',
                    field=self.fields['rel_related'],
                    label=_("Related"),
                    weight=20,
                    id_fields=[self.fields['id']],
                    directed=False,
                )
            )

    def init_default_badges(self, exclude=None):
        """
        Initialize a set of default `BadgeSpec` instances.

        These rely on `FieldSpec` instances, which must have been added beforehand.
        """
        pass

    def add_scope(self, scope):
        self.scopes[scope.key] = scope

    def remove_scope(self, key):
        del self.scopes[key]

    def add_field(self, field):
        self.fields[field.key] = field
        if field.field_type:
            self.schema.add(field.key, field.field_type)

    def remove_field(self, key):
        self.schema.remove(key)
        del self.fields[key]

    def add_facet(self, facet):
        self.facets[facet.key] = facet
        self.schema.add(facet.key, facet.field_type)

    def remove_facet(self, key):
        self.schema.remove(key)
        del self.facets[key]

    def get_facet_by_filter_key(self, filter_key, default=None):
        for spec in self.facets.values():
            if spec.filter_key == filter_key:
                return spec
        return default

    def add_sort(self, sort):
        self.sorts[sort.key] = sort

    def remove_sort(self, key):
        del self.sorts[key]

    def add_citation_format(self, citation_format):
        self.citation_formats[citation_format.key] = citation_format

    def remove_citation_format(self, key):
        del self.citation_formats[key]

    def add_badge(self, badge):
        self.badges[badge.key] = badge

    def remove_badge(self, key):
        del self.badges[key]

    def add_relation(self, relation):
        self.relations[relation.key] = relation

    def remove_relation(self, key):
        del self.relations[key]

    def get_ordered_specs(self, attr):
        """
        Return a list of specifications, sorted by weight.

        :param str attr: Attribute name of the specification dict. The
            specifications must themselves have a `weight` attribute.
        """
        return sorted(getattr(self, attr).values(), key=lambda spec: spec.weight)



"""
Dedicated container to display current search criteria.

The name "breadbox" is a reference to another well-known navigational aid
user interface pattern: the "breadcrumbs".
"""

from collections import defaultdict
from copy import copy

from flask import current_app


def build_breadbox_keywords(criteria):
    """Build the active keywords in the same format as a facet."""
    keywords = defaultdict(list)
    for scope in current_app.config['KERKO_COMPOSER'].scopes.values():
        if scope.key in criteria.keywords:
            for value in criteria.keywords.getlist(scope.key):
                keywords[scope.key].append({
                    'label': value,
                    'remove_url': criteria.build_remove_keywords_url(scope.key, value),
                })
    return keywords


def get_active_filters(results):
    filters = []
    for item in results:
        if item.get('remove_url', None):
            if item.get('children', []):
                item = copy(item)
                item['children'] = get_active_filters(item['children'])
            filters.append(item)
    return filters


def build_breadbox_filters(facet_results):
    """Collect just the active facets from the given facet results."""
    filters = {}
    for key, results in facet_results.items():
        facet_filters = get_active_filters(results)
        if facet_filters:
            filters[key] = facet_filters
    return filters


def build_breadbox(criteria, facet_results):
    breadbox = {}

    breadbox_keywords = build_breadbox_keywords(criteria)
    if breadbox_keywords:
        breadbox['keywords'] = breadbox_keywords

    breadbox_filters = build_breadbox_filters(facet_results)
    if breadbox_filters:
        breadbox['filters'] = breadbox_filters

    return breadbox



from flask import current_app


def build_sorter(criteria):
    composer = current_app.config['KERKO_COMPOSER']
    sorter = {
        'active': composer.sorts[criteria.sort].label,
        'items': [],
    }
    for sort_spec in composer.get_ordered_specs('sorts'):
        if sort_spec.is_allowed(criteria):
            sorter['items'].append({
                'label': sort_spec.label,
                'url': criteria.build_url(sort=sort_spec.key)
            })
    return sorter



import re
from collections.abc import Iterable

from flask import current_app
from flask_babel import gettext
from whoosh.qparser import MultifieldParser, QueryParser, plugins
from whoosh.query import And, Every, Not, Or, Term
from whoosh.sorting import Count, Facets, FieldFacet

from .criteria import Criteria
from .index import open_index


def get_search_return_fields(page_len, exclude=None):
    if exclude is None:
        exclude = []
    if page_len != 1:  # Note: page_len can be None (for all results).
        return_fields = [f for f in current_app.config['KERKO_RESULTS_FIELDS'] if f not in exclude]
        for badge in current_app.config['KERKO_COMPOSER'].badges.values():
            if badge.field.key not in return_fields:
                return_fields.append(badge.field.key)
        return return_fields
    return None  # All fields.


def build_keywords_query(keywords):
    """
    Build parsers for a query.

    :param MultiDict keywords: The search texts keyed by scope key. If empty,
        the query will match every documents.
    """
    queries = []
    if keywords:
        composer = current_app.config['KERKO_COMPOSER']
        text_plugins = [
            plugins.PhrasePlugin(),
            plugins.GroupPlugin(),
            plugins.OperatorsPlugin(
                And=r"(?<=\s)" + re.escape(gettext("AND")) + r"(?=\s)",
                Or=r"(?<=\s)" + re.escape(gettext("OR")) + r"(?=\s)",
                Not=r"(^|(?<=(\s|[()])))" + re.escape(gettext("NOT")) + r"(?=\s)",
                AndNot=None,
                AndMaybe=None,
                Require=None
            ),
            plugins.BoostPlugin(),
        ]
        for key, value in keywords.items(multi=True):
            fields = [spec.key for spec in composer.fields.values() if key in spec.scopes]
            if not fields:
                raise KeyError  # No known field for that scope key.
            parser = MultifieldParser(
                fields, schema=composer.schema, plugins=text_plugins
            )
            queries.append(parser.parse(value))
    else:
        queries.append(Every())
    return And(queries)


def get_query_facets(filters, criteria):
    """Return the specs of facets specified by the search criteria."""
    composer = current_app.config['KERKO_COMPOSER']
    if criteria.page_len == 1:
        # On single-item pages, facets are displayed in the breadbox only, thus
        # only the active facets need to be retrieved.
        facets = []
        if filters:
            for filter_key, _ in filters:
                spec = composer.get_facet_by_filter_key(filter_key)
                if spec:
                    facets.append(spec)
        return facets
    return composer.facets.values()


def build_groupedby_query(facets):
    """
    Build the faceting part of a search query.

    :param list facets: A list of `FacetSpec` objects.
    """
    groupedby = Facets()
    for spec in facets:
        groupedby.add_field(spec.key, allow_overlap=spec.allow_overlap)
    return groupedby


def build_filter_query(filters=None):
    """
    Build the filtering part of a search query.

    :param list filters: A list of (name, values) tuples, where `values` is
        itself a list.
    """
    composer = current_app.config['KERKO_COMPOSER']
    terms = []
    if filters:
        for filter_key, filter_values in filters:
            spec = composer.get_facet_by_filter_key(filter_key)
            if spec:  # Ensure only valid filters.
                for v in filter_values:
                    if v == '':  # If trying to filter with a missing value.
                        # Exclude all results with a value in facet field.
                        terms.append(Not(Every(spec.key)))
                    else:
                        v = spec.codec.transform_for_query(v)
                        terms.append(spec.query_class(spec.key, v))
    return And(terms)


def _get_fields(hit, return_fields=None):
    """
    Copy fields from a Hit object into a dict.

    Fields must be copied while the Searcher object is alive.

    :param whoosh.searching.Hit hit: The search result.

    :param list return_fields: A list of fields to load. If None, all fields
        will be loaded.
    """
    if return_fields is None:
        item = hit.fields()
    else:
        item = {f: hit[f] for f in return_fields if f in hit}
    return _decode_fields(item)


def _decode_fields(item):
    field_specs = current_app.config['KERKO_COMPOSER'].fields
    for k, v in item.items():
        if k in field_specs:
            item[k] = field_specs[k].decode(v)
    return item


def build_creators_display(item):
    """Prepare creator data for easier display."""
    for spec in item.keys():
        if spec == 'data' and 'creators' in item['data']:
            for creator in item['data']['creators']:
                # Add creator display name.
                creator['display'] = creator.get('name', '') or ', '.join(
                    [
                        n for n in (
                            creator.get('lastName', ''),
                            creator.get('firstName', '')
                        ) if n
                    ]
                )
                # Add creator type labels.
                if 'creator_types' in item:
                    for t in item['creator_types']:
                        if t['creatorType'] == creator['creatorType']:
                            creator['label'] = t['localized']
                            break
                # Add creator link.
                if 'creator' in current_app.config['KERKO_COMPOSER'].scopes:
                    creator['url'] = Criteria().build_add_keywords_url(
                        scope='creator',
                        value='"{}"'.format(creator['display'])
                    )


def build_item_facet_results(item):
    """
    Prepare facet "results" describing the item.

    The only distinction from real facet results obtained from a search is that
    counts are unknown. Using the same data structure as real facet results
    allows the reuse of facet display logic.
    """
    item['facet_results'] = {}
    for spec in current_app.config['KERKO_COMPOSER'].facets.values():
        if spec.key in item:
            if isinstance(item[spec.key], list):
                fake_results = [(value, 0) for value in item[spec.key]]
            else:
                fake_results = [(item[spec.key], 0)]
            # Pass an empty Criteria; the facets will provide starting points for new searches.
            item['facet_results'][spec.key] = spec.build(fake_results, criteria=Criteria())


def build_search_facet_results(searcher, groups, criteria, facet_specs):
    """
    Prepare facet results for the search page.
    """
    facets = {}
    if groups:
        # Build facet results from groupings obtained with the search.
        for spec in facet_specs:
            facets[spec.key] = spec.build(groups(spec.key).items(), criteria)
    elif criteria.has_filter_search():
        # No groupings available even though facets are used. This usually means
        # that the search itself had zero results, thus no facet results either.
        # But building facet results is still desirable in order to display the
        # active filters in the search interface. To get those, we perform a
        # separate query for each active filter, but this time ignoring any
        # other search criteria.
        for filter_key in criteria.filters.keys():
            for spec in facet_specs:
                if filter_key == spec.filter_key:
                    results = searcher.search(
                        Every(),
                        filter=build_filter_query(
                            [tuple([spec.key, criteria.filters.getlist(spec.key)])]
                        ),
                        groupedby=build_groupedby_query([spec]),
                        maptype=Count,  # Not to be used, as other criteria are ignored.
                        limit=1,  # Don't care about the documents.
                    )
                    facets[spec.key] = spec.build(
                        results.groups(spec.key).items(), criteria, active_only=True
                    )
    return facets


def _get_directed_relation_search_terms(item, relation):
    """
    Build search terms for a directed relation.

    Finds items for which one of the identifiers match an identifier from the
    current item's relation field.
    """
    return [
        Term(field.key, ref) for field in relation.id_fields
        for ref in item.get(relation.field.key, [])
    ]


def _get_reverse_relation_search_terms(item, relation):
    """
    Build search terms for a reverse relation.

    Finds occurrences of one of the identifiers of the current item in the
    relation field of other items.
    """
    identifiers = []
    for id_field in relation.id_fields:
        value = item.get(id_field.key, [])
        if isinstance(value, str):
            identifiers.append(value)
        else:
            identifiers.extend(value)
    return [Term(relation.field.key, identifier) for identifier in identifiers]


def build_relations(item, return_fields=None, sort=None):
    """
    Prepare the relational fields of the item for a given relation.
    """
    index = open_index()
    if index:
        with index.searcher() as searcher:
            composer = current_app.config['KERKO_COMPOSER']
            if sort in composer.sorts:
                search_args = build_sort_args(composer.sorts[sort])
            else:
                search_args = {}

            def search_relation(search_terms, item, key):
                """
                Search for a relation and store the result in the given item (at key).
                """
                results = searcher.search(q=Or(search_terms), limit=None, **search_args)
                if results:
                    item[key] = [
                        _get_fields(hit, return_fields) for hit in results
                    ]
                else:
                    item[key] = []  # Replace original value of the field with the results.

            for relation in composer.get_ordered_specs('relations'):
                if relation.directed:
                    search_relation(
                        search_terms=_get_directed_relation_search_terms(item, relation),
                        item=item,
                        key=relation.field.key,
                    )
                else:
                    search_relation(
                        search_terms=(
                            _get_directed_relation_search_terms(item, relation) +
                            _get_reverse_relation_search_terms(item, relation)
                        ),
                        item=item,
                        key=relation.field.key,
                    )
                if relation.reverse:
                    search_relation(
                        search_terms=_get_reverse_relation_search_terms(item, relation),
                        item=item,
                        key=relation.reverse_field_key,
                    )


def build_sort_args(sort_spec):
    if not sort_spec.fields:
        return {}
    if isinstance(sort_spec.reverse, Iterable):
        # Per-field reverse values require to use FieldFacet objects.
        return {
            'sortedby':
                [
                    FieldFacet(field_key, reverse=reverse)
                    for field_key, reverse in zip(sort_spec.get_field_keys(), sort_spec.reverse)
                ]
        }
    # With a single reverse value, we may specify the field names directly.
    return {
        'sortedby': sort_spec.get_field_keys(),
        'reverse': sort_spec.reverse,
    }


def run_query(criteria, return_fields=None, query_facets=True):
    """Perform a search query."""
    items = []
    facets = {}
    total = 0
    page_count = 0
    last_sync = None

    index = open_index()
    if index:
        last_sync = index.last_modified()
        with index.searcher() as searcher:
            composer = current_app.config['KERKO_COMPOSER']
            q = build_keywords_query(criteria.keywords)
            search_args = {
                'filter': build_filter_query(criteria.filters.lists()),
            }
            search_args.update(build_sort_args(composer.sorts[criteria.sort]))
            if query_facets:
                facet_specs = get_query_facets(criteria.filters.lists(), criteria)
                search_args['groupedby'] = build_groupedby_query(facet_specs)
                search_args['maptype'] = Count
            groups = None
            if criteria.page_len is None:  # Retrieve all results.
                search_args['limit'] = None

                results = searcher.search(q, **search_args)
                if results:
                    items = [_get_fields(hit, return_fields) for hit in results]
                    groups = results.groups
                    total = results.estimated_length()
                    page_count = 1
            else:  # Retrieve a range of results.
                search_args['pagenum'] = criteria.page_num
                search_args['pagelen'] = criteria.page_len

                results = searcher.search_page(q, **search_args)
                if results:
                    items = [_get_fields(hit, return_fields) for hit in results]
                    groups = results.results.groups
                    total = results.total
                    page_count = results.pagecount
            if query_facets:
                facets = build_search_facet_results(searcher, groups, criteria, facet_specs)

    return items, facets, total, page_count, last_sync


def run_query_unique(field_name, value, return_fields=None):
    """Perform a search query for a single item using an unique key."""
    index = open_index()
    if index:
        with index.searcher() as searcher:
            q = QueryParser(
                field_name,
                schema=current_app.config['KERKO_COMPOSER'].schema,
                plugins=[]
            ).parse(value)
            results = searcher.search(q, limit=1)
            if results:
                return _get_fields(results[0], return_fields)
    return None


def run_query_unique_with_fallback(field_names, value, return_fields=None):
    """
    Query multiple fields for a single item using an unique key.

    :return: A tuple with the found item (or `None` if not found), and the index
        of the field that contained a match (or `None` if not found).
    """
    for i, field_name in enumerate(field_names):
        result = run_query_unique(field_name, value, return_fields)
        if result:
            return result, i
    return None, None


def run_query_all(return_fields=None):
    """Perform a search query to return all items (without faceting)."""
    index = open_index()
    if index:
        with index.searcher() as searcher:
            results = searcher.search(Every(), limit=None)
            if results:
                for hit in results:
                    yield _get_fields(hit, return_fields)
    return []


def check_fields(fields):
    """
    Check if the specified fields exist in the schema.

    :param list fields: List if field names to check.

    :return list: List of fields missing from the schema, if any.
    """
    index = open_index()
    if index:
        return [name for name in fields if name not in index.schema.names()]
    return fields



from abc import ABC, abstractmethod
from collections.abc import Iterable

from babel.numbers import format_number
from flask_babel import get_locale
from whoosh.fields import ID
from whoosh.query import Prefix, Term

from . import extractors, renderers
from .codecs import BaseFacetCodec, CollectionFacetCodec, IdentityFieldCodec
from .text import slugify, sort_normalize
from .tree import Tree


class ScopeSpec:
    """
    Specifies a scope for keyword searches.

    This is a configuration element, with no effect on the search index schema.
    """

    def __init__(self, key, selector_label, breadbox_label, weight=0, help_text=''):
        self.key = key
        self.selector_label = selector_label
        self.breadbox_label = breadbox_label
        self.weight = weight
        self.help_text = help_text


class BaseFieldSpec(ABC):

    def __init__(
            self,
            key,
            field_type,
            extractor,
    ):
        """
        Initialize this field specification.

        :param str key: Unique key for referencing the field in the search index
            schema.

        :param whoosh.fields.FieldType field_type: Instance of the schema field
            type. Set to `None` if the field's value is to be passed along with
            documents, but not added to the schema, e.g., boost factors (see
            `whoosh.writing.IndexWriter.add_document`).

        :param kerko.extractors.Extractor extractor: Instance of the extractor
            that will extract the field value from a Zotero item.
        """
        self.key = key
        self.field_type = field_type
        self.extractor = extractor

    def extract_to_document(self, document, item_context, library_context):
        """Extract the value of this element from a Zotero item."""
        self.extractor.extract_and_store(document, item_context, library_context, self)


class FieldSpec(BaseFieldSpec):
    """Specifies a schema field."""

    def __init__(
            self,
            codec=None,
            scopes=None,
            **kwargs
    ):
        """
        Initialize this field specification.

        :param BaseFieldCodec codec: Value encoder/decoder for this field.

        :param list scopes: List of keys to `ScopeSpec` instances to which this
            field applies. If `None`, the field won't be available for keyword
            search.
        """
        super().__init__(**kwargs)
        self.scopes = scopes or []
        self.codec = codec or IdentityFieldCodec()

    def encode(self, value):
        return self.codec.encode(value)

    def decode(self, encoded_value):
        return self.codec.decode(encoded_value)


class FacetSpec(BaseFieldSpec):
    """Specifies a facet for search grouping and filtering."""

    def __init__(
            self,
            title,
            filter_key,
            weight=0,
            codec=None,
            missing_label=None,
            sort_key=None,
            sort_reverse=False,
            item_view=True,
            allow_overlap=True,
            query_class=None,
            collapsible=True,
            renderer=None,
            **kwargs
    ):
        """
        Initialize this facet specification.

        :param str title: Title of the facet.

        :param str filter_key: Key to use in URLs when filtering with this facet.

        :param int weight: Determine the position of this facet relative to the
            others.

        :param BaseFacetCodec codec: Value encoder/decoder for this facet.

        :param bool item_view: Show this facet on item view pages.

        :param `renderers.Renderer` renderer: A renderer for this facet. The
            rendering context provides the following variables:
            - `spec`: The `FacetSpec` instance.
            - `items`: The facet values retrieved by the search query.
            - `mode`: A string whose value is one of the following:
                - `'search'`: Displaying the facet on a search results page.
                - `'field'`: Displaying the facet on a full bibliographic
                  record page.
                - `'breadbox'`: Displaying the facet as a search criteria in
                  the breadbox.

        .. seealso: Additional :meth:`BaseFieldSpec.__init__` arguments.
        """
        super().__init__(**kwargs)
        self.title = title
        self.filter_key = filter_key
        self.weight = weight
        self.codec = codec or BaseFacetCodec()
        self.missing_label = missing_label
        self.sort_key = sort_key
        self.sort_reverse = sort_reverse
        self.item_view = item_view
        self.allow_overlap = allow_overlap
        self.query_class = query_class or Term
        self.collapsible = collapsible
        self.renderer = renderer or renderers.TemplateResolverRenderer(
            'kerko/_facet_{mode}.html.jinja2'
        )

    def encode(self, value):
        return self.codec.encode(value)

    def decode(self, encoded_value, default_value=None, default_label=None):
        if encoded_value is None:
            return '', ''
        return self.codec.decode(encoded_value, default_value, default_label)

    @abstractmethod
    def add_filter(self, value, active_filters):
        """
        Build a query string that adds a value for this facet to the active filters.

        :param string value: The value to add to the active filters.

        :param MultiDict active_filters: The active filters to derive from.

        :return MultiDict: The resulting filters.
        """

    @abstractmethod
    def remove_filter(self, value, active_filters):
        """
        Build a query string that removes a value for this facet from the active filters.

        :param string value: The value to remove from the active filters.

        :param MultiDict active_filters: The active filters to derive from.

        :return MultiDict: The resulting filters.
        """

    @abstractmethod
    def build(self, results, criteria, active_only=False):
        """
        Construct a facet's items for display.

        :param iterable results: Iterable of (value, count) tuples representing
            the facet's results from the search query.

        :param Criteria criteria: The current search criteria. If None, the
            facet will be built for performing a new search.

        :param active_only: Only build the items that are related to active
            filters, i.e., filters actually present in the search criteria.
        """

    def sort_items(self, items):
        if self.sort_key is None:
            return
        # Sort items with multiple-keys, using tuples. The first tuple element
        # (the primary sort key), gets inserted before the other sort keys. It
        # is a boolean value that ensures that empty labels (those of missing
        # values) appear last regardless of the sort direction.
        #
        # Note: Count values are multiplied by -1 to reverse their order.
        # Because there is only one 'reverse' flag for all sort keys, the
        # reverse order for counts is, counterintuitively, ascending. When
        # ordering facet values by count, we usually want descending order, so
        # if the reverse flag False in that case, it is possible to get the
        # second sort key, usually 'label' sorted in alphabetical order.
        return sorted(
            items,
            key=lambda x: (
                bool(x['label']) if self.sort_reverse else not x['label'], *[
                    x[k] * -1 if k == 'count' else
                    (sort_normalize(x[k]) if isinstance(x[k], str) else x[k])
                    for k in self.sort_key
                ]
            ),
            reverse=self.sort_reverse
        )

    def render(self, items, mode):
        return self.renderer.render(spec=self, items=items, mode=mode)


class FlatFacetSpec(FacetSpec):

    def add_filter(self, value, active_filters):
        if value is None:  # Special case for missing value (None is returned by Whoosh).
            value = ''
        filters = active_filters.deepcopy()
        active_values = filters.getlist(self.filter_key)
        if active_values and value in active_values:
            return None  # Already filtering with value. No add needed.
        filters.setlistdefault(self.filter_key).append(value)
        return filters

    def remove_filter(self, value, active_filters):
        if value is None:  # Special case for missing value (None is returned by Whoosh).
            value = ''
        filters = active_filters.deepcopy()
        active_values = filters.getlist(self.filter_key)
        if not active_values or value not in active_values:
            return None  # Not currently filtering with value. No remove needed.
        active_values.remove(value)
        filters.setlist(self.filter_key, active_values)
        return filters

    def build(self, results, criteria, active_only=False):
        items = []
        for value, count in results:
            if value or self.missing_label:
                value, label = self.decode(value, default_value=value, default_label=value)
                remove_url = criteria.build_remove_filter_url(self, value)
                if remove_url or active_only:
                    add_url = None
                else:
                    add_url = criteria.build_add_filter_url(self, value)
                if remove_url or add_url:  # Only items with an URL get displayed.
                    items.append({
                        'label': label,
                        'count': count,
                        'count_formatted': format_number(count, locale=get_locale()),
                        'remove_url': remove_url,
                        'add_url': add_url,
                    })
        return self.sort_items(items)


class TreeFacetSpec(FacetSpec):

    def __init__(self, path_separator='.', **kwargs):
        super().__init__(**kwargs)
        self.path_separator = path_separator

    @staticmethod
    def is_ancestor(ancestor, descendant):
        """
        Return True if `ancestor` is an ancestor of `descendant`.

        :param str ancestor: Potential ancestor.

        :param str descendant: Path to look for potential ancestor.
        """
        # True if ancestor is a prefix of descendant. Warning: This simple
        # condition only works for finding ancestors when all path components
        # have the same length, otherwise any partial prefix could match.
        return descendant.find(ancestor) == 0 and ancestor != descendant

    def get_parent(self, value):
        parent = value.rsplit(sep=self.path_separator, maxsplit=1)[0]
        return parent if parent != value else None

    def add_filter(self, value, active_filters):
        if value is None:
            # Special case for missing value (None is returned by Whoosh).
            value = ''
        filters = active_filters.deepcopy()
        active_values = filters.getlist(self.filter_key)
        for i, active_value in enumerate(active_values):
            if value == active_value or self.is_ancestor(value, active_value):
                # Already filtering with value or a descendant. No add needed.
                return None
            elif self.is_ancestor(active_value, value):
                # Active value is ancestor of value. Replace the active value.
                active_values[i] = value
                filters.setlist(self.filter_key, active_values)
                break
        else:
            # This is an all new filter. Add its value.
            filters.setlistdefault(self.filter_key).append(value)
        return filters

    def remove_filter(self, value, active_filters):
        if value is None:
            # Special case for missing value (None is returned by Whoosh).
            value = ''
        filters = active_filters.deepcopy()
        active_values = filters.getlist(self.filter_key)
        if not active_values:
            # Filter not active at all. No remove needed.
            return None
        new_values = []
        change = False
        for active_value in active_values:
            if value == active_value or self.is_ancestor(value, active_value):
                # Currently filtering with value or a descendant. Remove needed.
                change = True
                parent = self.get_parent(value)
                if parent and parent not in new_values:
                    new_values.append(parent)
            else:
                # Filter unrelated to value. Preserve it.
                new_values.append(active_value)
        if not change:
            # Neither filtering with value or a descendant. No remove needed.
            return None
        filters.setlist(self.filter_key, new_values)
        return filters

    def sort_tree(self, tree):
        """
        Convert tree to list, sorting children along the way.

        In the process, each item remains a dict as in a flat facet, but with
        an extra 'children' key.
        """
        lst = []
        for child in tree['children'].values():
            child['node']['children'] = self.sort_tree(child)
            lst.append(child['node'])
        return self.sort_items(lst)

    def build(self, results, criteria, active_only=False):
        tree = Tree()
        for value, count in results:
            if value or self.missing_label:
                value, label = self.decode(value, default_value=value, default_label=value)
                remove_url = criteria.build_remove_filter_url(self, value)
                if remove_url or active_only:
                    add_url = None
                else:
                    add_url = criteria.build_add_filter_url(self, value)
                if remove_url or add_url:  # Only items with an URL get displayed.
                    if value:
                        path = value.split(sep=self.path_separator)
                    else:
                        path = [value]

                    # Build the tree path. Part of the path may or may not already
                    # exist as the facet values are not ordered.
                    node = tree  # Start at tree root.
                    for component in path:
                        node = node['children'][component]
                    # Add data at the leaf.
                    node['node'] = {
                        'label': label,
                        'count': count,
                        'count_formatted': format_number(count, locale=get_locale()),
                        'remove_url': remove_url,
                        'add_url': add_url,
                    }
        return self.sort_tree(tree)


class CollectionFacetSpec(TreeFacetSpec):
    """
    Specifies a facet based on a top-level Zotero collection.

    A top-level Zotero collection can act as a facet when its key matches a
    given `collection_key`. Subcollections become values within the facet.
    """

    def __init__(self, collection_key, **kwargs):
        # Provide more convenient defaults for this type of facet.
        kwargs.setdefault('key', 'collection_facet_' + collection_key)
        kwargs.setdefault('field_type', ID(stored=True))
        kwargs.setdefault('filter_key', slugify(kwargs.get('title')))
        kwargs.setdefault('codec', CollectionFacetCodec())
        kwargs.setdefault('sort_key', ['label'])
        kwargs.setdefault('query_class', Prefix)
        kwargs.setdefault('extractor', extractors.CollectionFacetTreeExtractor())
        super().__init__(**kwargs)
        self.collection_key = collection_key


class SortSpec:
    """
    Specifies a sort option.

    This is a configuration element, with no effect on the search index schema.
    """

    def __init__(
            self,
            key,
            label,
            fields,
            weight=0,
            reverse=False,
            is_allowed=True
    ):
        """
        Initialize a sort option.

        :param str key: Key of this sort option.

        :param str label: Label of this sort option.

        :param list fields: List of `FieldSpec` instances to use when doing
            search queries with this sort option, in order of precedence.

        :param int weight: Determine the position of this option relative to the
            other options.

        :param bool reverse: Whether the fields should be sorted in reverse
            order. To provide per-field reverse settings, an iterable may be
            supplied instead of a bool, in which case it should contain the same
            number of elements as the `fields` parameter.

        :param callable is_allowed: Optional callable to determine if, given a
            criteria object, the sort option should be allowed.
        """
        self.key = key
        self.label = label
        self.fields = fields
        self.weight = weight
        if isinstance(reverse, Iterable):
            if all(reverse):
                self.reverse = True
            elif not any(reverse):
                self.reverse = False
            else:
                self.reverse = reverse
        else:
            self.reverse = reverse
        self._is_allowed = is_allowed

    def is_allowed(self, criteria):
        if callable(self._is_allowed):
            return self._is_allowed(criteria)
        return self._is_allowed

    def get_field_keys(self):
        if self.fields:
            return [spec.key for spec in self.fields]
        return None


class CitationFormatSpec:
    """
    Specifies a citation download format.

    This is a configuration element, with no effect on the search index schema.
    """

    def __init__(
            self,
            key,
            field,
            label,
            help_text,
            weight,
            extension,
            mime_type,
            group_format='{}',
            group_item_delimiter=''
    ):
        """
        Initialize a citation download format.

        :param str key: Key of this citation format.

        :param FieldSpec field: `FieldSpec` instance associated to this citation
            format.

        :param str label: Label of this citation format.

        :param int weight: Determine the position of this citation format
            relative to the others in lists.

        :param str extension: File extension of this citation format.

        :param str mime_type: MIME type of this citation format.

        :param str group_format: Format string for wrapping multiple entries.

        :param str group_item_delimiter: Delimiter string to insert between
            entries.
        """
        self.key = key
        self.field = field
        self.label = label
        self.help_text = help_text
        self.weight = weight
        self.extension = extension
        self.mime_type = mime_type
        self.group_format = group_format
        self.group_item_delimiter = group_item_delimiter


class RelationSpec:
    """
    Specifies a type of relation between items.

    This is a configuration element, with no effect on the search index schema.
    """

    def __init__(
            self,
            *,
            key,
            field,
            label,
            weight,
            id_fields,
            directed=True,
            reverse=False,
            reverse_key='',
            reverse_field_key='',
            reverse_label=''
    ):
        """
        Initialize a relation type.

        :param str key: Key of this relation type.

        :param FieldSpec field: `FieldSpec` instance associated to this relation
            type. The field will be looked up in the search index to retrieve
            item identifiers related to a given item.

        :param str label: Label of this relation type.

        :param int weight: Determine the position of this relation type
            relatively to other relation types in lists.

        :paras list id_fields: List of `FieldSpec` instances representing the
            fields to search when trying to resolve an item identifier.

        :param bool directed: Whether a relation is directed (i.e.
            bidirectional) or not.

        :param bool reverse: Whether a reverse relation should be exposed. If
            `directed` is `False`, this can only be `False` as well.

        :param str reverse_key: Key of the reverse relation. Should be set only
            if `reverse` is `True`.

        :param str reverse_field_key: Field key to use for storing the reverse
            relation. This isn't a `FieldSpec` as the field won't be looked up
            in the search index. Instead, it will be dynamically populated with
            items whose `field` contain a given item. Should be set only if
            `reverse` is `True`.

        :param str reverse_label: Label of the reverse relation. Should be set
            only if `reverse` is `True`.
        """
        assert not reverse and not directed or directed
        assert not reverse and not reverse_key or reverse
        assert not reverse and not reverse_field_key or reverse
        assert not reverse and not reverse_label or reverse
        self.key = key
        self.field = field
        self.label = label
        self.weight = weight
        self.id_fields = id_fields
        self.directed = directed
        self.reverse = reverse
        self.reverse_key = reverse_key
        self.reverse_field_key = reverse_field_key
        self.reverse_label = reverse_label


class BadgeSpec:
    """
    Specifies a badge.

    Badges may be displayed on items (in search results and on full
    bibliographic record pages).

    This is a configuration element, with no effect on the search index schema.
    """

    def __init__(
            self,
            key,
            field,
            activator,
            renderer,
            weight=0,
    ):
        """
        Initialize this badge specification.

        :param str key: Key of this badge.

        :param FieldSpec field: `FieldSpec` instance required by this badge.

        :param callable activator: Callable which, given a `FieldSpec` instance
            and an item, must return `True` if the badge should be displayed.

        :param `renderers.Renderer` renderer: A renderer for this badge. The
            rendering context provides the following variables:
            - `field`: The `FieldSpec` instance.
            - `item`: The item retrieved from the search index.
            - `mode`: A string whose value is one of the following:
                - `'result':` The item is being viewed in a list of results.
                - `'item'`: Viewing the item's full bibliographic record.

        :param int weight: Determine the position of this badge relative to the
            others.
        """
        self.key = key
        self.field = field
        self.activator = activator
        self.renderer = renderer
        self.weight = weight

    def is_active(self, item):
        """
        Return `True` is this badge is active for the given item.
        """
        return self.activator(self.field, item)

    def render(self, item, mode):
        """
        Render the badge, if necessary, for the given item.

        :return str: The rendered badge, or `''` if the badge is not activated
            on the item.
        """
        if self.is_active(item):
            return self.renderer.render(field=self.field, item=item, mode=mode)
        return ''



from wtforms import SelectField, validators
from wtforms.fields.html5 import SearchField

from flask import current_app
from flask_wtf import FlaskForm


class SearchForm(FlaskForm):
    scope = SelectField()
    keywords = SearchField(validators=[validators.optional(), validators.length(max=1000)])

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.scope.choices = [
            (s.key, s.selector_label)
            for s in current_app.config['KERKO_COMPOSER'].get_ordered_specs('scopes')
        ]



"""
Kerko: A Flask blueprint that provides faceted search for bibliographies based
on Zotero.
"""
# pylint: disable=invalid-name

import pathlib

from flask import Blueprint
from flask_babel import Domain

from .jinja2 import register_filters


babel_domain = Domain(domain='kerko')


def init_default_config(state):
    """
    Initialize default KERKO settings in the app's config object.

    The following settings must be set in the app's config and have no defaults:

    - `KERKO_ZOTERO_LIBRARY_ID`
    - `KERKO_ZOTERO_LIBRARY_TYPE`
    - `KERKO_ZOTERO_API_KEY`
    - `KERKO_DATA_DIR`

    The following settings are used with Bootstrap-Flask to load CDN-based
    resources. Alternatively, one may set `BOOTSTRAP_SERVE_LOCAL=True` to use
    the static resources included with the Bootstrap-Flask extension (and ignore
    these settings):

    - `KERKO_BOOTSTRAP_VERSION`
    - `KERKO_JQUERY_VERSION`
    - `KERKO_POPPER_VERSION`
    - `KERKO_WITH_JQUERY`
    - `KERKO_WITH_POPPER`
    """
    state.app.config.setdefault('KERKO_TITLE', 'Kerko')
    state.app.config.setdefault('KERKO_WHOOSH_LANGUAGE', 'en')
    state.app.config.setdefault('KERKO_ZOTERO_LOCALE', 'en-US')
    state.app.config.setdefault('KERKO_ZOTERO_START', 0)
    state.app.config.setdefault('KERKO_ZOTERO_END', 0)
    state.app.config.setdefault('KERKO_ZOTERO_MAX_ATTEMPTS', 10)
    state.app.config.setdefault('KERKO_ZOTERO_WAIT', 120)  # In seconds.
    state.app.config.setdefault('KERKO_ZOTERO_BATCH_SIZE', 100)
    state.app.config.setdefault('KERKO_PAGE_LEN', 20)
    state.app.config.setdefault('KERKO_PAGER_LINKS', 4)
    state.app.config.setdefault('KERKO_CSL_STYLE', 'apa')
    state.app.config.setdefault('KERKO_RESULTS_ABSTRACTS', False)
    state.app.config.setdefault('KERKO_RESULTS_ABSTRACTS_TOGGLER', True)
    state.app.config.setdefault(
        'KERKO_RESULTS_FIELDS', ['id', 'attachments', 'bib', 'coins', 'data']
    )
    state.app.config.setdefault('KERKO_FACET_COLLAPSING', False)
    state.app.config.setdefault('KERKO_PRINT_ITEM_LINK', False)
    state.app.config.setdefault('KERKO_PRINT_CITATIONS_LINK', False)
    state.app.config.setdefault('KERKO_PRINT_CITATIONS_MAX_COUNT', 0)
    state.app.config.setdefault('KERKO_DOWNLOAD_CITATIONS_LINK', True)
    state.app.config.setdefault('KERKO_DOWNLOAD_CITATIONS_MAX_COUNT', 0)
    state.app.config.setdefault('KERKO_DOWNLOAD_ATTACHMENT_NEW_WINDOW', False)
    state.app.config.setdefault('KERKO_RELATIONS_INITIAL_LIMIT', 5)
    state.app.config.setdefault('KERKO_RELATIONS_SORT', 'author_asc')
    state.app.config.setdefault('KERKO_USE_TRANSLATIONS', True)

    state.app.config.setdefault('KERKO_TEMPLATE_BASE', 'kerko/base.html.jinja2')
    state.app.config.setdefault('KERKO_TEMPLATE_LAYOUT', 'kerko/layout.html.jinja2')
    state.app.config.setdefault('KERKO_TEMPLATE_SEARCH', 'kerko/search.html.jinja2')
    state.app.config.setdefault('KERKO_TEMPLATE_SEARCH_ITEM', 'kerko/search-item.html.jinja2')
    state.app.config.setdefault('KERKO_TEMPLATE_ITEM', 'kerko/item.html.jinja2')

    state.app.config.setdefault('KERKO_BOOTSTRAP_VERSION', '4.5.3')
    state.app.config.setdefault('KERKO_JQUERY_VERSION', '3.5.1')
    state.app.config.setdefault('KERKO_POPPER_VERSION', '1.16.1')
    state.app.config.setdefault('KERKO_WITH_JQUERY', True)
    state.app.config.setdefault('KERKO_WITH_POPPER', True)

    # Look for translations both in Kerko's and the app's translations directories.
    state.app.config.setdefault(
        'BABEL_TRANSLATION_DIRECTORIES',
        ';'.join([str(pathlib.Path(__file__).parent / 'translations'), 'translations'])
    )


blueprint = Blueprint(
    'kerko', __name__, static_folder='static', template_folder='templates'
)
blueprint.record_once(init_default_config)
register_filters(blueprint)


from . import views  # pylint: disable=wrong-import-position



"""
Functions for extracting data from Zotero items.
"""

import itertools
import re
from abc import ABC, abstractmethod
from collections.abc import Iterable

from flask import Markup, current_app

from .tags import TagGate
from .text import id_normalize, sort_normalize
from .transformers import find_item_id_in_zotero_uris_str

RECORD_SEPARATOR = '\x1e'


class ItemContext:
    """Contains data related to a Zotero item."""

    def __init__(self, item, children):
        """
        Initialize an item's data.

        :param dict item: An item, as returned by the Zotero API.

        :param list children: A list of dicts representing the item's children.
        """
        self.item_key = item.get('key', '')  # For convenient access.
        self.item = item
        self.children = children
        self.data = item.get('data', {})  # For convenient access.


class LibraryContext:
    """Contains data related to a Zotero library."""

    def __init__(
            self, library_id, library_type, *, collections, item_types, item_fields, creator_types
    ):
        self.library_id = library_id
        self.library_type = library_type
        self.collections = collections
        self.item_types = item_types
        self.item_fields = item_fields
        self.creator_types = creator_types

    def get_creator_types(self, item_data):
        return self.creator_types.get(item_data.get('itemType', ''), [])


def _parse_zotero_date(text):
    """Parse a fuzzy date into a (year, month, day) tuple of numerical values."""
    year = month = day = 0
    matches = re.match(r'^([0-9]{4})(-([0-9]{2})(-([0-9]{2}))?)?', text)
    if matches:
        if matches.group(1):
            year = int(matches.group(1))
        if matches.group(3):
            month = int(matches.group(3))
        if matches.group(5):
            day = int(matches.group(5))
    return year, month, day


def encode_single(value, spec):
    """Encode a single value."""
    return spec.encode(value)


def encode_multiple(value, spec):
    """Encode items of an iterable."""
    return [spec.encode(item) for item in value]


class Extractor(ABC):
    """
    Data extractor.

    An extractor can retrieve elements from ``ItemContext`` or `LibraryContext``
    objects and add them to a document. The document is represented by a `dict`.
    A ``BaseFieldSpec`` object provides both an ``encode()`` method that may
    transform the data before its assignment into the document, and the key to
    assign the resulting data to.
    """

    def __init__(self, format_='data', encode=encode_single, **kwargs):
        """
        Initialize the extractor.

        :param str format_: Format to retrieve when performing the Zotero item
            read, e.g., 'data', 'bib', 'ris', to ensure that the data required
            by this extractor is available in the ``ItemContext`` object at
            extraction time.

        :param callable encode: Function that can encode a value using a
            ``FieldSpec``.
        """
        self.format = format_
        self.encode = encode
        assert not kwargs  # Subclasses should have consumed every keyword arg.

    @abstractmethod
    def extract(self, item_context, library_context, spec):
        """
        Retrieve the value from context.

        :return: Extracted value, or `None` if no value could be extracted.
        """

    def extract_and_store(self, document, item_context, library_context, spec):
        """
        Extract value from context and store its encoded version in document.
        """
        extracted_value = self.extract(item_context, library_context, spec)
        if extracted_value is not None:
            document[spec.key] = self.encode(extracted_value, spec)

    def warning(self, message, item_context):
        current_app.logger.warning(
            f"{self.__class__.__name__}: {message} ({item_context.item_key})"
        )


class TransformerExtractor(Extractor):
    """
    Wrap an extractor to transform data before encoding it into the document.
    """

    def __init__(self, *, extractor, transformers, **kwargs):
        """
        Initialize the extractor.

        :param Extractor extractor: Base extractor to wrap.

        :param list transformers: List of callables that will be chained to
            transform the extracted data. Each callable takes a value as
            argument and returns the transformed value.
        """
        super().__init__(format_=extractor.format, **kwargs)
        self.extractor = extractor
        self.transformers = transformers

    def apply_transformers(self, value):
        for transformer in self.transformers:
            value = transformer(value)
        return value

    def extract(self, item_context, library_context, spec):
        value = self.extractor.extract(item_context, library_context, spec)
        return self.apply_transformers(value)


class MultiExtractor(Extractor):
    """
    Allow a composition of multiple extractors.
    """

    def __init__(self, *, extractors, encode=encode_multiple, **kwargs):
        super().__init__(encode=encode, **kwargs)
        self.extractors = extractors

    def extract(self, item_context, library_context, spec):
        values = []
        for extractor in self.extractors:
            assert self.format == extractor.format  # Extractors can only use same format as parent.
            value = extractor.extract(item_context, library_context, spec)
            if isinstance(value, Iterable) and not isinstance(value, str):
                values.extend(value)
            elif value:
                values.append(value)
        return values or None


class KeyExtractor(Extractor):  # pylint: disable=abstract-method

    def __init__(self, *, key, **kwargs):
        """
        Initialize the extractor.

        :param str key: Key of the element to extract from the Zotero item.
        """
        super().__init__(**kwargs)
        self.key = key


class ItemExtractor(KeyExtractor):
    """Extract a value from an item."""

    def extract(self, item_context, library_context, spec):
        return item_context.item.get(self.key)


class ItemDataExtractor(KeyExtractor):
    """Extract a value from item data."""

    def extract(self, item_context, library_context, spec):
        return item_context.data.get(self.key)


class RawDataExtractor(Extractor):

    def extract(self, item_context, library_context, spec):
        return item_context.data


class ItemRelationsExtractor(Extractor):
    """Extract a list of item's relations corresponding to a given predicate."""

    def __init__(self, predicate, **kwargs):
        super().__init__(**kwargs)
        self.predicate = predicate

    def extract(self, item_context, library_context, spec):
        relations = item_context.data.get('relations', {}).get(self.predicate, [])
        if relations and isinstance(relations, str):
            relations = [relations]
        assert isinstance(relations, Iterable)
        return relations


class ItemTypeLabelExtractor(Extractor):
    """Extract the label of the item's type."""

    def extract(self, item_context, library_context, spec):
        item_type = item_context.data.get('itemType')
        if item_type and item_type in library_context.item_types:
            return library_context.item_types[item_type]
        self.warning("Missing itemType", item_context)
        return None


class ItemFieldsExtractor(Extractor):
    """Extract field metadata, serialized as a JSON string."""

    def extract(self, item_context, library_context, spec):
        item_type = item_context.data.get('itemType')
        if item_type and item_type in library_context.item_fields:
            fields = library_context.item_fields[item_type]
            # Retain metadata for fields that are actually present in the item.
            item_fields = [f for f in fields if f.get('field') in item_context.data]
            if item_fields:
                return item_fields
            self.warning("Missing item type fields", item_context)
        else:
            self.warning("Missing itemType", item_context)
        return None


class CreatorTypesExtractor(Extractor):
    """Extract creator types metadata, serialized as a JSON string."""

    def extract(self, item_context, library_context, spec):
        item_type = item_context.data.get('itemType')
        if item_type and item_type in library_context.creator_types:
            library_creator_types = library_context.creator_types[item_type]
            # Retain metadata for creator types that are actually present in the item.
            item_creator_types = []
            for library_creator_type in library_creator_types:
                for item_creator in item_context.data.get('creators', []):
                    creator_type = item_creator.get('creatorType')
                    if creator_type and creator_type == library_creator_type.get('creatorType'):
                        item_creator_types.append(library_creator_type)
                        break
            if item_creator_types:
                return item_creator_types
            if item_context.data.get('creators', False):
                self.warning("Missing creator types", item_context)
        else:
            self.warning("Missing itemType", item_context)
        return None


class CreatorsExtractor(Extractor):
    """Flatten and extract creator data."""

    def extract(self, item_context, library_context, spec):
        creators = []
        if 'creators' in item_context.data:
            for creator in item_context.data['creators']:
                n = creator.get('name', '').strip()
                if n:
                    creators.append(n)
                firstname = creator.get('firstName', '').strip()
                lastname = creator.get('lastName', '').strip()
                if firstname and lastname:
                    # Combine firstname and lastname in different orders to help
                    # phrase searches.
                    creators.append(' '.join([firstname, lastname]))
                    creators.append(', '.join([lastname, firstname]))
                elif firstname:
                    creators.append(firstname)
                elif lastname:
                    creators.append(lastname)
        return RECORD_SEPARATOR.join(creators) if creators else None


class CollectionNamesExtractor(Extractor):
    """Extract item collections for text search."""

    def extract(self, item_context, library_context, spec):
        names = set()
        if 'collections' in item_context.data:
            for k in item_context.data['collections']:
                if k in library_context.collections:
                    name = library_context.collections[k].get('data', {}).get('name', '').strip()
                    if name:
                        names.add(name)
        return RECORD_SEPARATOR.join(names) if names else None


class BaseTagsExtractor(Extractor):

    def __init__(self, *, include_re='', exclude_re='', **kwargs):
        """
        Initialize the extractor.

        :param str include_re: Any tag that does not matches this regular
            expression will be ignored by the extractor. If empty, all tags will
            be accepted unless `exclude_re` is set and they match it.

        :param str exclude_re: Any tag that matches this regular expression
            will be ignored by the extractor. If empty, all tags will be
            accepted unless `include_re` is set and they do not match it.
        """
        super().__init__(**kwargs)
        self.include = re.compile(include_re) if include_re else None
        self.exclude = re.compile(exclude_re) if exclude_re else None

    def extract(self, item_context, library_context, spec):
        tags = set()
        if 'tags' in item_context.data:
            for tag_data in item_context.data['tags']:
                tag = tag_data.get('tag', '').strip()
                if tag and \
                        (not self.include or self.include.match(tag)) and \
                        (not self.exclude or not self.exclude.match(tag)):
                    tags.add(tag)
        return tags or None


class TagsTextExtractor(BaseTagsExtractor):
    """Extract item tags for text search."""

    def extract(self, item_context, library_context, spec):
        tags = super().extract(item_context, library_context, spec)
        return RECORD_SEPARATOR.join(tags) if tags else None


class BaseChildrenExtractor(Extractor):

    def __init__(self, *, item_type, include_re='', exclude_re='', **kwargs):
        """
        Initialize the extractor.

        :param str item_type: The type of child items to extract, either 'note'
            or 'attachment'.

        :param str include_re: Any child which does not have a tag that
            matches this regular expression will be ignored by the extractor. If
            empty, all children will be accepted unless `exclude_re` is set
            and causes some to be rejected.

        :param str exclude_re: Any child that have a tag that matches this
            regular expression will be ignored by the extractor. If empty, all
            children will be accepted unless `include_re` is set and causes
            some to be rejected.
        """
        super().__init__(**kwargs)
        self.item_type = item_type
        self.gate = TagGate(include_re, exclude_re)

    def extract(self, item_context, library_context, spec):
        accepted_children = []
        for child in item_context.children:
            if child.get('data', {}).get('itemType') == self.item_type \
                    and self.gate.check(child.get('data', {})):
                accepted_children.append(child)
        return accepted_children or None


class BaseAttachmentsExtractor(BaseChildrenExtractor):

    def __init__(self, **kwargs):
        super().__init__(item_type='attachment', **kwargs)


class StoredFileAttachmentsExtractor(BaseAttachmentsExtractor):
    """
    Extract the metadata of stored copies of files into a list of dicts.
    """

    def __init__(self, *, mime_types=None, **kwargs):
        super().__init__(**kwargs)
        self.mime_types = mime_types

    def is_attachment(self, child):
        if not child.get('data'):
            return False
        if not child.get('key'):
            return False
        if child['data'].get('linkMode') != 'imported_file':
            return False
        if self.mime_types and child['data'].get(
                'contentType', 'octet-stream'
        ) not in self.mime_types:
            return False
        return True

    def extract(self, item_context, library_context, spec):
        children = super().extract(item_context, library_context, spec)
        return [
            {
                'id': child['key'],
                'mimetype': child['data'].get('contentType', 'octet-stream'),
                'filename': child['data'].get('filename', child['key']),
                'md5': child['data'].get('md5', ''),
                'mtime': child['data'].get('mtime', 0),
            } for child in children if self.is_attachment(child)
        ] if children else None


class LinkedURIAttachmentsExtractor(BaseAttachmentsExtractor):
    """
    Extract attached links to URIs into a list of dicts.
    """

    @staticmethod
    def is_link(child):
        if not child.get('data'):
            return False
        if not child.get('key'):
            return False
        if child['data'].get('linkMode') != 'linked_url':
            return False
        if not child['data'].get('url'):
            return False
        return True

    def extract(self, item_context, library_context, spec):
        children = super().extract(item_context, library_context, spec)
        return [
            {
                'title': child['data'].get('title', child['data'].get('url')),
                'url': child['data'].get('url')
            }
            for child in children if self.is_link(child)
        ] if children else None


class BaseNotesExtractor(BaseChildrenExtractor):  # pylint: disable=abstract-method

    def __init__(self, **kwargs):
        super().__init__(item_type='note', **kwargs)


class NotesTextExtractor(BaseNotesExtractor):
    """Extract notes for text search."""

    def extract(self, item_context, library_context, spec):
        children = super().extract(item_context, library_context, spec)
        return RECORD_SEPARATOR.join(
            [
                Markup(child.get('data', {}).get('note', '')).striptags()
                for child in children
            ]
        ) if children else None


class RawNotesExtractor(BaseNotesExtractor):
    """Extract raw notes for storage."""

    def extract(self, item_context, library_context, spec):
        children = super().extract(item_context, library_context, spec)
        return [child.get('data', {}).get('note', '') for child in children] if children else None


class RelationsInNotesExtractor(BaseNotesExtractor):
    """Extract item references specified in child notes."""

    def extract(self, item_context, library_context, spec):
        refs = []
        children = super().extract(item_context, library_context, spec)
        if children:
            for child in children:
                note = child.get('data', {}).get('note', '')
                note = Markup(re.sub(r'<br\s*/>', '\n', note)).striptags()  # Strip HTML markup.
                refs.extend(find_item_id_in_zotero_uris_str(note))
        return refs or None


def _expand_paths(path):
    """
    Extract the paths of each of the components of the specified path.

    If the given path is ['a', 'b', 'c'], the returned list of paths is:
    [['a'], ['a', 'b'], ['a', 'b', 'c']]
    """
    return [path[0:i + 1] for i in range(len(path))]


class CollectionFacetTreeExtractor(Extractor):
    """Index the Zotero item's collections needed for the specified facet."""

    def __init__(self, encode=encode_multiple, **kwargs):
        super().__init__(encode=encode, **kwargs)

    def extract(self, item_context, library_context, spec):
        # Sets prevent duplication when multiple collections share common ancestors.
        encoded_ancestors = set()
        for collection_key in item_context.data.get('collections', []):
            if collection_key not in library_context.collections:
                continue  # Skip unknown collection.
            ancestors = library_context.collections.ancestors(collection_key)
            if len(ancestors) <= 1 or ancestors[0] != spec.collection_key:
                continue  # Skip collection, unrelated to this facet.

            ancestors = ancestors[1:]  # Facet values come from subcollections.
            for path in _expand_paths(ancestors):
                label = library_context.collections.get(
                    path[-1], {}
                ).get('data', {}).get('name', '').strip()
                encoded_ancestors.add((tuple(path), label))  # Cast path to make it hashable.
        return encoded_ancestors or None


class InCollectionExtractor(Extractor):
    """Extract the boolean membership of an item into a collection."""

    def __init__(self, *, collection_key, true_only=True, check_subcollections=True, **kwargs):
        """
        Initialize the extractor.

        :param str collection_key: Key of the collection to test item membership
            against.

        :param bool true_only: If `True` (default), extraction returns `True`
            when an item belongs to the specified collection, or `None` when it
            does not belong to that collection. If `False`, always return a
            boolean.

        :param bool check_subcollections: If `True` (default), membership is
            extended to any subcollection of the specified collection.
        """
        super().__init__(**kwargs)
        self.collection_key = collection_key
        self.true_only = true_only
        self.check_subcollections = check_subcollections

    def extract(self, item_context, library_context, spec):
        item_collections = list(
            itertools.chain(
                *[
                    library_context.collections.ancestors(c) if self.check_subcollections else c
                    for c in item_context.data.get('collections', [])
                ]
            )
        )
        is_in = self.collection_key in item_collections
        if not self.true_only:
            return is_in
        if is_in:
            return True
        return None


class TagsFacetExtractor(BaseTagsExtractor):
    """Index the Zotero item's tags for faceting."""

    def __init__(self, encode=encode_multiple, **kwargs):
        super().__init__(encode=encode, **kwargs)


class ItemTypeFacetExtractor(Extractor):
    """Index the Zotero item's type for faceting."""

    def extract(self, item_context, library_context, spec):
        item_type = item_context.data.get('itemType')
        if item_type:
            return (item_type, library_context.item_types.get(item_type, item_type))
        self.warning("Missing itemType", item_context)
        return None


class YearFacetExtractor(Extractor):
    """Index the Zotero item's publication date for faceting by year."""

    def __init__(self, encode=encode_multiple, **kwargs):
        super().__init__(encode=encode, **kwargs)

    def extract(self, item_context, library_context, spec):
        parsed_date = item_context.item.get('meta', {}).get('parsedDate', '')
        if parsed_date:
            year, _month, _day = _parse_zotero_date(parsed_date)
            decade = int(int(year) / 10) * 10
            century = int(int(year) / 100) * 100
            return _expand_paths([str(century), str(decade), str(year)])
        return None


class ItemDataLinkFacetExtractor(ItemDataExtractor):

    def extract(self, item_context, library_context, spec):
        return item_context.data.get(self.key, '').strip() != ''


def _prepare_sort_text(text):
    """
    Normalize the given text for a sort field.

    Sort fields are bytearrays in the schema, so the text goes through
    str.encode().

    :param str text: The Unicode string to normalize.

    :return bytearray: The normalized text.
    """
    return sort_normalize(Markup(text).striptags()).encode()


class SortItemDataExtractor(ItemDataExtractor):

    def extract(self, item_context, library_context, spec):
        return _prepare_sort_text(item_context.data.get(self.key, ''))


class SortCreatorExtractor(Extractor):

    def extract(self, item_context, library_context, spec):
        creators = []

        def append_creator(creator):
            creator_parts = [
                _prepare_sort_text(creator.get('lastName', '')),
                _prepare_sort_text(creator.get('firstName', '')),
                _prepare_sort_text(creator.get('name', ''))]
            creators.append(b' zzz '.join([p for p in creator_parts if p]))

        # We treat creator types like an ordered list, where the first creator
        # type is for primary creators. Depending on the citation style, lesser
        # creator types may not appear in citations. Therefore, we try to sort
        # only by primary creators in order to avoid sorting with data that may
        # be invisible to the user. Only when an item has no primary creator do
        # we fallback to lesser creators.
        for creator_type in library_context.get_creator_types(item_context.data):
            for creator in item_context.data.get('creators', []):
                if creator.get('creatorType', '') == creator_type.get('creatorType'):
                    append_creator(creator)
            if creators:
                break  # No need to include lesser creator types.
        return b' zzzzzz '.join(creators)


class SortDateExtractor(Extractor):

    def extract(self, item_context, library_context, spec):
        parsed_date = item_context.item.get('meta', {}).get('parsedDate', '')
        year, month, day = _parse_zotero_date(parsed_date)
        return int('{:04d}{:02d}{:02d}'.format(year, month, day))
